import numpy as np
from itertools import product

from ._lattice_edge_logic import get_naive_edges, create_padded_sites


# Function to get N distinct colors
def get_n_colors(N):
    import matplotlib.pyplot as plt

    # Use 'tab10' colormap for distinct colors
    colors = plt.get_cmap("tab10", N)
    return [colors(i) for i in range(N)]


def default_arguments(dim, node_size, node_text_color, node_text_offset):
    """
    Compute default arguments when unspecified
    """
    if node_size is None:
        if dim == 1 or dim == 2:
            node_size = 300
        else:
            node_size = 50
    if node_text_offset is None:
        if dim == 1:
            node_text_offset = 0
        elif dim == 2:
            node_text_offset = 0
        elif dim == 3:
            node_text_offset = 0.05
    if node_text_color is None:
        if node_text_offset == 0:
            node_text_color = "white"
        else:
            node_text_color = "black"

    if isinstance(node_text_offset, (float, int)):
        node_text_offset = np.full((dim,), node_text_offset)
    return node_size, node_text_color, node_text_offset


def ax_text_autodim(
    ax,
    position,
    label,
    *,
    offset: list | float = 0,
    **kwargs,
):
    """
    Equivalent to ax.text but automatically handles
    1D/2D/3D positions, with an optional offset.
    """
    dim = len(position)
    if dim < 1 or dim > 3:
        raise NotImplementedError("...")

    if isinstance(offset, (float, int)):
        offset = np.full((dim,), 0.0)
    if isinstance(offset, (list, tuple)):
        offset = np.array(offset)

    if dim == 1:
        # 1D is still a 2D plot, the offset is along y
        dpos = np.array([position[0], offset[0]])
    else:
        dpos = position + offset

    return ax.text(*dpos, label, **kwargs)


def ax_scatter_autodim(
    ax,
    positions,
    *args,
    **kwargs,
):
    """
    Equivalent to ax.text but automatically handles
    1D/2D/3D positions, with an optional offset.
    """
    dim = positions.shape[-1]
    if dim < 1 or dim > 3:
        raise NotImplementedError("...")

    if positions.ndim == 1:
        positions = np.atleast_2d(positions)

    if dim == 1:
        # 1D is still a 2D plot, the offset is along y
        pos = [positions[:, 0], np.zeros_like(positions[:, 0])]
    elif dim == 2:
        pos = [positions[:, 0], positions[:, 1]]
    elif dim == 3:
        pos = [positions[:, 0], positions[:, 1], positions[:, 2]]

    return ax.scatter(
        *pos,
        *args,
        **kwargs,
    )


def draw_lattice(
    self,
    ax=None,
    figsize: tuple[int | float] | None = None,
    distance_order=1,
    *,
    node_size: int | None = None,
    node_color: str = "#1f78b4",
    node_text_color: str | None = None,
    node_text_offset: float | None = None,
    extra_sites_alpha: float = 0.3,
    draw_neighbors: bool = True,
    draw_text: bool = True,
    draw_basis_vectors: bool = True,
    draw_unit_cell: bool = True,
    show: bool = True,
    **kwargs,
):
    """
    Draws the ``Lattice`` graph

    Args:
        ax: A Matplotlib axis object. If unspecified it creates a figure.
        figsize: (width, height) tuple of the generated figure, if no axes is
            specified.
        node_size: Size of the nodes (as in matplotlib.pyplot.scatter).
        node_color: String with the colour of the nodes.
        node_text_color: String with the colour of the integer labelling the nodes.
            Defaults to white for 1D/2D and black in 3D.
        node_text_offset: float or 1/2/3-vector of offset for the label from the node.
            Defaults to 0 for 1/2D and to 0.05 for 3D.
        extra_sites_alpha: A float value between 0 and 1 for the alpha of the sites
            outside of the extent. Defaults to 0.3
        extra_sites_alpha: If true, we also draw sites outside of the extent
            of the lattice corresponding to the periodic images. (Defaults True)
        draw_text: If True, we draw the integer label with the site index (defaults to True).
        draw_basis_vectors: If True, draw the basis vectors (defaults to True).
        draw_unit_cell: If True, shade an area corresponding to the unit cell starting
            at the origin (Defaults to True).
        show: If True, show the plot before returning it (defaults to True).

    Returns:
        Matplotlib axis object containing the graph's drawing.
    """
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection

    lattice = self

    # Determine the dimensionality of the lattice
    basis_vectors = lattice.basis_vectors
    dim = basis_vectors.shape[0]

    if dim < 1 or dim > 3:
        raise TypeError("Cannot draw 0D or 4D lattices")

    if len(kwargs.keys()) > 0:
        for deprecated_key in ("edge_color", "curvature", "font_size", "font_color"):
            if deprecated_key in kwargs.keys():
                kwargs.pop(deprecated_key)
                print(
                    f"Keyword argument {deprecated_key} is deprecated and does nothing anymore."
                )
        if len(kwargs.keys()) > 0:
            raise NotImplementedError(f"unsupported kwargument {tuple(kwargs.keys())}")

    node_size, node_text_color, node_text_offset = default_arguments(
        dim, node_size, node_text_color, node_text_offset
    )

    # extent = lattice.extent
    distance_atol = 1e-5

    # Set up the plot
    if ax is None:
        if dim == 1 or dim == 2:
            fig, ax = plt.subplots(figsize=figsize)
        else:  # 3D
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111, projection="3d")

    # Calculate all site positions
    all_sites = []
    for cell in product(*[range(e) for e in lattice.extent]):
        for offset in lattice.site_offsets:
            position = np.dot(cell, basis_vectors) + offset
            all_sites.append(position)
    all_sites = np.array(all_sites)

    all_sites, ids = create_padded_sites(
        basis_vectors, lattice.extent, lattice.site_offsets, lattice.pbc, 0
    )
    if draw_neighbors:
        all_sites_extra, ids_extra = create_padded_sites(
            basis_vectors,
            lattice.extent,
            lattice.site_offsets,
            lattice.pbc,
            distance_order,
        )

        naive_edges_by_order = get_naive_edges(
            all_sites_extra,
            distance_order * np.linalg.norm(basis_vectors, axis=1).max()
            + distance_atol,
            distance_order,
        )

        # remove virtual vertices that are not connected to physical ones
        union_vertices_to_delete = []
        to_delete = set()
        to_keep = set()
        for k, naive_edges in enumerate(naive_edges_by_order):
            for node1, node2 in naive_edges:
                # switch to real node indices
                pos1 = all_sites_extra[node1]
                pos2 = all_sites_extra[node2]

                # check if not within lattice
                pos1_id = lattice.positions[ids_extra[node1]]
                pos2_id = lattice.positions[ids_extra[node2]]

                if not (all(pos1_id == pos1) or all(pos2_id == pos2)):
                    to_delete.add(node1)
                    to_delete.add(node2)
                    # print(f"should delete {node1}, {node2} ({ids_extra[node1]},{ids_extra[node2]}) : ")
                    # print(f" -> {pos1_id =}, {pos1 = }, {pos2_id = }, {pos2 = }")
                else:
                    to_keep.add(node1)
                    to_keep.add(node2)
            # union_vertices_to_delete.append(vertices_to_delete)
        # union_vertices_to_delete = set(*union_vertices_to_delete)
        to_delete.difference_update(to_keep)
        union_vertices_to_delete = np.array(list(to_delete), dtype=np.int32)

        # for i in sorted(union_vertices_to_delete, reverse=True):
        #   print("deleting i", all_sites_extra[i])
        #   del all_sites_extra[i]
        #   del ids_extra[i]
        all_sites_extra = np.delete(all_sites_extra, union_vertices_to_delete, axis=0)
        ids_extra = np.delete(ids_extra, union_vertices_to_delete)

        # rgenerea
        naive_edges_by_order = get_naive_edges(
            all_sites_extra,
            distance_order * np.linalg.norm(basis_vectors, axis=1).max()
            + distance_atol,
            distance_order,
        )

    else:
        all_sites_extra = all_sites
        ids_extra = ids

        naive_edges_by_order = get_naive_edges(
            all_sites,
            distance_order * np.linalg.norm(basis_vectors, axis=1).max()
            + distance_atol,
            distance_order,
        )

    # Plot edges
    colors = get_n_colors(len(naive_edges_by_order))
    for k, naive_edges in enumerate(naive_edges_by_order):
        a = 1 - k * 0.2
        color = colors[k]
        for node1, node2 in naive_edges:
            # switch to real node indices
            pos1 = all_sites_extra[node1]
            pos2 = all_sites_extra[node2]

            # check if not within lattice
            pos1_id = lattice.positions[ids_extra[node1]]
            pos2_id = lattice.positions[ids_extra[node2]]

            if all(pos1_id == pos1) and all(pos2_id == pos2):
                linestyle = "-"
            else:
                linestyle = "--"

            if dim == 1:
                ax.plot(
                    [pos1[0], pos2[0]],
                    [0, 0],
                    linestyle=linestyle,
                    alpha=a,
                    color=color,
                )
            elif dim == 2:
                ax.plot(
                    [pos1[0], pos2[0]],
                    [pos1[1], pos2[1]],
                    linestyle=linestyle,
                    alpha=a,
                    color=color,
                )
            else:  # 3D
                ax.plot(
                    [pos1[0], pos2[0]],
                    [pos1[1], pos2[1]],
                    [pos1[2], pos2[2]],
                    linestyle=linestyle,
                    alpha=a,
                    color=color,
                )

    # Plot sites
    zorders = [
        artist.zorder for artist in ax.get_children() if hasattr(artist, "zorder")
    ]
    zorder = max(zorders) + 1

    if draw_neighbors:
        ax_scatter_autodim(
            ax,
            all_sites_extra,
            s=node_size,
            c=node_color,
            alpha=extra_sites_alpha,
            zorder=zorder,
        )
    ax_scatter_autodim(
        ax,
        all_sites,
        s=node_size,
        c=node_color,
        zorder=zorder + 1,
    )

    if draw_text:
        for pos, site in zip(all_sites_extra, ids_extra):
            ax_text_autodim(
                ax,
                pos,
                str(site),
                offset=node_text_offset,
                fontsize=8,
                horizontalalignment="center",
                verticalalignment="center",
                zorder=zorder + 2,
                color=node_text_color,
            )

    if draw_basis_vectors:
        # Plot basis vectors as arrows
        origin = np.zeros(dim)
        for i, basis in enumerate(basis_vectors):
            labl = f"$\\vec{{b}}_{{{i}}}$"
            if dim == 2:
                ax.quiver(
                    origin[0],
                    origin[1],
                    basis[0],
                    basis[1],
                    angles="xy",
                    scale_units="xy",
                    scale=1,
                    color=f"C{i}",
                    width=0.005,
                    headwidth=10,
                    headlength=10,
                    zorder=-10,
                )
                midpoint = basis / 2
                ax.text(
                    midpoint[0],
                    midpoint[1],
                    labl,
                    fontsize=10,
                    ha="center",
                    va="center",
                    # bbox=dict(facecolor="white", edgecolor="none", alpha=0.7),
                )
            elif dim == 3:  # 3D
                ax.quiver(
                    origin[0],
                    origin[1],
                    origin[2],
                    basis[0],
                    basis[1],
                    basis[2],
                    color=f"C{i}",
                    zorder=-10,
                )
                midpoint = basis / 2
                ax.text(midpoint[0], midpoint[1], midpoint[2], labl, fontsize=10)

    if draw_unit_cell:
        origin = np.zeros(dim)
        if dim == 2:
            # Shade the parallelogram formed by the basis vectors
            parallelogram = np.array(
                [
                    origin,
                    basis_vectors[0],
                    basis_vectors[0] + basis_vectors[1],
                    basis_vectors[1],
                ]
            )
            ax.add_patch(
                plt.Polygon(parallelogram, alpha=0.2, facecolor="gray", zorder=-20),
            )

        elif dim == 3:
            # Shade the parallelpiped formed by the basis vectors
            vertices = np.array(
                [
                    origin,
                    basis_vectors[0],
                    basis_vectors[1],
                    basis_vectors[2],
                    basis_vectors[0] + basis_vectors[1],
                    basis_vectors[0] + basis_vectors[2],
                    basis_vectors[1] + basis_vectors[2],
                    basis_vectors[0] + basis_vectors[1] + basis_vectors[2],
                ]
            )

            faces = [
                [vertices[0], vertices[1], vertices[4], vertices[2]],
                [vertices[0], vertices[1], vertices[5], vertices[3]],
                [vertices[0], vertices[2], vertices[6], vertices[3]],
                [vertices[7], vertices[4], vertices[1], vertices[5]],
                [vertices[7], vertices[4], vertices[2], vertices[6]],
                [vertices[7], vertices[5], vertices[3], vertices[6]],
            ]

            ax.add_collection3d(Poly3DCollection(faces, alpha=0.2, facecolor="gray"))

    # Set labels and title
    ax.set_xlabel("X")
    if dim >= 2:
        ax.set_ylabel("Y")
    if dim == 3:
        ax.set_zlabel("Z")
    ax.set_title(f"{dim}D Lattice (Distance Order: {distance_order})")

    if dim == 1:
        ax.set_ylim(-0.2, 0.2)
    elif dim == 2:
        ax.set_aspect("equal", "box")
    else:  # 3D
        # Equal aspect ratio for 3D plots
        ax.set_box_aspect((1, 1, 1))

    # Show the plot
    if show:
        plt.show()

    return ax



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from collections.abc import Callable
from collections.abc import Hashable

from netket.jax import canonicalize_dtypes
from netket.utils.types import DType, PyTree, Array

from netket.hilbert import AbstractHilbert
from netket.operator import AbstractOperator


class ContinuousOperator(AbstractOperator):
    r"""This class is the abstract base class for operators defined on a
    continuous Hilbert space. Users interested in implementing new
    quantum Operators for continuous Hilbert spaces should subclass
    `ContinuousOperator` and implement its interface.
    """

    def __init__(self, hilbert: AbstractHilbert, dtype: DType | None = None):
        r"""
        Constructs the continuous operator acting on the given hilbert space and
        with a certain data type.

        Args:
            hilbert: The underlying Hilbert space on which the operator is defined
            dtype: Data type of the operator, which is used to infer the dtype of
                expectation values
        """
        dtype = canonicalize_dtypes(float, dtype=dtype)
        self._dtype = dtype
        self._hash = None
        super().__init__(hilbert)

    @property
    def dtype(self) -> DType:
        return self._dtype

    @abc.abstractmethod
    def _expect_kernel(
        self, logpsi: Callable, params: PyTree, x: Array, data: PyTree | None
    ):
        r"""This method defines the action of the local operator on a given quantum state
        `logpsi` for a given configuration `x`.
        :math:`O_{loc}(x) =  \frac{\bra{x}O{\ket{\psi}}{\bra{x}\ket{\psi}}`
        This method is executed inside of a `jax.jit` block.
        Any static data from the operator itself should be captured in the method.
        Any additional data is provided by the `_pack_arguments`-method
        and will be passed as the `data` argument in this method (Example: masses in kinetic energy).

        Args:
            logpsi: variational state
            params: parameters for the variational state
            x: a sample of particle positions
            data: additional data
        """

    @abc.abstractmethod
    def _pack_arguments(self) -> PyTree | None:
        r"""This methods should return a PyTree that will be passed as the `data` argument
        to the `_expect_kernel`. The PyTree should be composed of jax arrays or hashable
        objects.

        For example for the kinetic energy this method would return the masses of the
        individual particles."""

    @abc.abstractproperty
    def _attrs(self) -> tuple[Hashable, ...]:
        """This must return a tuple of (hashable) attributes used to compare two operators of
        the same type and that is hashed to compute the hash of the operator.

        To hash arrays you can return them as `nk.utils.HashableArray`.
        """

    def __add__(self, other):
        if isinstance(self, ContinuousOperator) and isinstance(
            other, ContinuousOperator
        ):
            from netket.operator import SumOperator

            return SumOperator(self, other)
        else:
            return NotImplemented  # pragma: no cover

    def __rmul__(self, other):
        if isinstance(self, ContinuousOperator) and isinstance(other, float):
            return self * other
        else:
            return NotImplemented  # pragma: no cover

    def __mul__(self, other):
        if isinstance(self, ContinuousOperator) and isinstance(other, float):
            from netket.operator import SumOperator

            return SumOperator(self, coefficients=other)
        else:
            return NotImplemented  # pragma: no cover

    def __sub__(self, other):
        if isinstance(other, ContinuousOperator):
            return self + (-other)
        else:
            return NotImplemented  # pragma: no cover

    def __neg__(self):
        return -1.0 * self

    def __hash__(self):
        if self._hash is None:
            self._hash = hash((type(self), self._attrs))
        return self._hash

    def __eq__(self, other) -> bool:
        if isinstance(other, type(self)):
            return self._attrs == other._attrs
        return False



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
from jax import numpy as jnp
from flax import struct

from netket.utils.types import Scalar, PyTree
from netket.utils import mpi
from netket import jax as nkjax

from ..linear_operator import LinearOperator, Uninitialized

from .common import check_valid_vector_type


@struct.dataclass
class QGTJacobianDenseT(LinearOperator):
    """
    Semi-lazy representation of an S Matrix behaving like a linear operator.

    The matrix of gradients O is computed on initialisation, but not S,
    which can be computed by calling :code:`to_dense`.
    The details on how the ⟨S⟩⁻¹⟨F⟩ system is solved are contained in
    the field `sr`.
    """

    O: jnp.ndarray = Uninitialized
    """Gradients O_ij = ∂log ψ(σ_i)/∂p_j of the neural network
    for all samples σ_i at given values of the parameters p_j
    Average <O_j> subtracted for each parameter
    Divided through with sqrt(#samples) to normalise S matrix
    If scale is not None, columns normalised to unit norm
    """

    scale: jnp.ndarray | None = None
    """If not None, contains 2-norm of each column of the gradient matrix,
    i.e., the sqrt of the diagonal elements of the S matrix
    """

    mode: str = struct.field(pytree_node=False, default=Uninitialized)
    """Differentiation mode:
        - "real": for real-valued R->R and C->R Ansätze, splits the complex inputs
                  into real and imaginary part.
        - "complex": for complex-valued R->C and C->C Ansätze, splits the complex
                  inputs and outputs into real and imaginary part
        - "imag": For the imaginary part of the QGT. Solve in this case builds the
                  purely imaginary hermitian matrix to solve the linear system.
        - "holomorphic": for any Ansätze. Does not split complex values.
        - "auto": autoselect real or complex.
    """

    _in_solve: bool = struct.field(pytree_node=False, default=False)
    """Internal flag used to signal that we are inside the _solve method and matmul should
    not take apart into real and complex parts the other vector"""

    _params_structure: PyTree = struct.field(pytree_node=False, default=Uninitialized)

    @jax.jit
    def __matmul__(self, vec: PyTree | jnp.ndarray) -> PyTree | jnp.ndarray:
        if not hasattr(vec, "ndim") and not self._in_solve:
            check_valid_vector_type(self._params_structure, vec)

        # When we do matrix multiplication, we convert the input vector to
        # the dense format used by the QGTJacobian. If we are using
        vec, reassemble = convert_tree_to_dense_format(
            vec, self.mode, disable=self._in_solve
        )

        if self.scale is not None:
            vec = vec * self.scale

        result = mat_vec(vec, self.O, self.diag_shift, imag=(self.mode == "imag"))

        if self.scale is not None:
            result = result * self.scale

        return reassemble(result)

    @jax.jit
    def _solve(self, solve_fun, y: PyTree, *, x0: PyTree | None = None) -> PyTree:
        if not hasattr(y, "ndim"):
            check_valid_vector_type(self._params_structure, y)

        y, reassemble = convert_tree_to_dense_format(y, self.mode)

        if x0 is not None:
            x0, _ = convert_tree_to_dense_format(x0, self.mode)
            if self.scale is not None:
                x0 = x0 * self.scale

        if self.scale is not None:
            y = y / self.scale

        # to pass the object LinearOperator itself down
        # but avoid rescaling, we pass down an object with
        # scale = None
        unscaled_self = self.replace(scale=None, _in_solve=True)
        if self.mode == "imag":
            # If we want to solve the linear system IM(G)x=vec,
            # the G matrix is hermitian so its imaginary part is skew-simmetric
            # and has purely imaginary eigenvalues. This is extemely unstable
            # numerically, so we instead solve the problem
            # i Im(G) x = i Vec, such that the matrix becomes hermitian and the
            # solving algorithm are more stable.
            # Then, we take the real part of the solution.
            #
            # To multiply Im(G) by i, as Im(G) = Oₗ† Oᵣ , I can multiply
            # O by √i
            #
            unscaled_self = unscaled_self.replace(O=unscaled_self.O * jnp.sqrt(1j))
            out, info = solve_fun(unscaled_self, 1j * y, x0=x0)
            out = out.real
        else:
            out, info = solve_fun(unscaled_self, y, x0=x0)

        if self.scale is not None:
            out = out / self.scale

        return reassemble(out), info

    @jax.jit
    def to_dense(self) -> jnp.ndarray:
        """
        Convert the lazy matrix representation to a dense matrix representation.

        Returns:
            A dense matrix representation of this S matrix.
        """
        if self.scale is None:
            O = self.O
            diag = jnp.eye(self.O.shape[-1])
        else:
            O = self.O * self.scale[jnp.newaxis, :]
            diag = jnp.diag(self.scale**2)

        # concatenate samples with real/Imaginary dimension
        if self.mode == "imag":
            # Equivalent to Jr.T@Ji - Ji.T@Jr
            flip_sign = jnp.array([1, -1]).reshape(1, 2, 1)
            Ol = (flip_sign * O).reshape(-1, O.shape[-1])
            Or = jnp.flip(O, axis=1).reshape(-1, O.shape[-1])
            return mpi.mpi_sum_jax(Ol.T @ Or)[0] + self.diag_shift * diag
        else:
            # Equivalent to Jr.T@Jr + Ji.T@Ji
            O = O.reshape(-1, O.shape[-1])
            return mpi.mpi_sum_jax(O.conj().T @ O)[0] + self.diag_shift * diag

    def to_real_part(self) -> "QGTJacobianDenseT":
        """
        Returns the operator computing real part of the complex, non holomorphic QGT.

        The real part of the QGT is used in the Stochastic Reconfiguration (SR)
        algorithm as well as in the McLachlan Variational Principles used to simulate
        Quantum Dynamics.
        See Table 1 in `"Theory of Variational Quantum Simulation" by Yuan et al.
        <https://arxiv.org/pdf/1812.08767>`_ for more details.


        .. note::

            This function can only be called on the non-holomorphic QGT.

        .. note::

            The returned object is another :code:`QGTJacobian***` object.

            NetKet does not currently implement the *full* complex QGT object
            for non-holomorphic functions, instead you need to keep on hand
            the two separate terms obtained by calling
            :meth:`~netket.optimizer.qgt.QGTJacobianDenseT.to_real_part` and
            :meth:`~netket.optimizer.qgt.QGTJacobianDenseT.to_imag_part`.

        See also :meth:`~netket.optimizer.qgt.QGTJacobianDenseT.to_imag_part`.
        """
        if self.mode == "imag":
            return self.replace(mode="complex")
        elif self.mode == "complex":
            return self
        elif self.mode == "real":
            return self
        else:
            raise ValueError(
                "Can only convert to real part the imaginary part of the"
                "QGT, not the holomorphic or sign-less one."
            )

    def to_imag_part(self) -> "QGTJacobianDenseT":
        """
        Returns the operator computing imaginary part of the complex, non holomorphic QGT.

        The imaginary part of the QGT is necessary to implement the TDVP variational
        principle for the quantum dynamics.
        See Table 1 in `"Theory of Variational Quantum Simulation" by Yuan et al.
        <https://arxiv.org/pdf/1812.08767>`_ for more details.

        .. note::

            This function can only be called on the non-holomorphic QGTs
            of a complex-valued wave-function.

        .. note::

            The returned object is another :code:`QGTJacobian***` object.

            NetKet does not currently implement the *full* complex QGT object
            for non-holomorphic functions, instead you need to keep on hand
            the two separate terms obtained by calling
            :meth:`~netket.optimizer.qgt.QGTJacobianDenseT.to_real_part` and
            :meth:`~netket.optimizer.qgt.QGTJacobianDenseT.to_imag_part`.

        See also :meth:`~netket.optimizer.qgt.QGTJacobianDenseT.to_real_part`.
        """

        if self.mode == "complex":
            return self.replace(mode="imag")
        elif self.mode == "imag":
            return self
        else:
            raise ValueError(
                "Can only convert to imaginary part the real part of the"
                "QGT, not the holomorphic or sign-less one."
            )

    def __repr__(self):
        return (
            f"QGTJacobianDense(diag_shift={self.diag_shift}, "
            f"scale={self.scale}, mode={self.mode})"
        )


#################################################
#####           QGT internal Logic          #####
#################################################


def mat_vec(v: PyTree, O: PyTree, diag_shift: Scalar, imag: bool = False) -> PyTree:
    if not imag:
        # Matrix vector product of the (real part, or holomorphic) QGT matrix
        # with a vector. In the standard case, it does the multiplication equivalent
        # to J_r.T@(J_r@v_r) + J_i.T@(J_i@v_i) + diag_shift*v
        w = O @ v
        res = jnp.tensordot(w.conj(), O, axes=w.ndim).conj()
        return mpi.mpi_sum_jax(res)[0] + diag_shift * v
    else:
        # Matrix vector product of the imaginary part of the QGT matrix
        # with a vector. This is equivalent to
        # J_r.T@(J_i@v_i) - J_i.T@(J_r@v_r) + diag_shift*v

        Or = jnp.flip(O, axis=1).reshape(-1, O.shape[-1])
        w = Or @ v

        flip_sign = jnp.array([1, -1]).reshape(1, 2, 1)
        Ol = (flip_sign * O).reshape(-1, O.shape[-1])
        res = jnp.tensordot(w.conj(), Ol, axes=w.ndim).conj()
        return mpi.mpi_sum_jax(res)[0] + diag_shift * v


def convert_tree_to_dense_format(vec, mode, *, disable=False):
    """
    Converts an arbitrary PyTree/vector which might be real/complex
    to the dense-(maybe-real)-vector used for QGTJacobian.

    The format is dictated by the sequence of operations chosen by
    `nk.jax.jacobian(..., dense=True)`. As `nk.jax.jacobian` first
    converts the pytree of parameters to real and then concatenates
    real and imaginary terms with a tree_ravel, we must do the same
    in here.
    """
    unravel = lambda x: x
    reassemble = lambda x: x
    if not disable:
        if mode != "holomorphic":
            vec, reassemble = nkjax.tree_to_real(vec)
        if not hasattr(vec, "ndim"):
            vec, unravel = nkjax.tree_ravel(vec)

    return vec, lambda x: reassemble(unravel(x))



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
import jax
from functools import partial

from netket.vqs import (
    MCState,
    expect,
    expect_and_grad,
    get_local_kernel,
    get_local_kernel_arguments,
)
import netket.jax as nkjax
from netket.utils import mpi

from .variance_operator import VarianceObservable


@expect.dispatch
def expect(vstate: MCState, variance_operator: VarianceObservable, chunk_size: None):
    if variance_operator.hilbert != vstate.hilbert:
        raise TypeError("Hilbert spaces should match")

    local_kernel = get_local_kernel(vstate, variance_operator.operator)
    local_kernel2 = get_local_kernel(vstate, variance_operator.operator_squared)

    sigma, args = get_local_kernel_arguments(vstate, variance_operator.operator)
    sigma, args2 = get_local_kernel_arguments(
        vstate, variance_operator.operator_squared
    )

    return expect_and_grad_inner_mc(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        local_kernel,
        local_kernel2,
        sigma,
        args,
        args2,
        return_grad=False,
    )


@expect_and_grad.dispatch
def expect_and_grad(
    vstate: MCState,
    variance_operator: VarianceObservable,
    chunk_size: None,
    *,
    mutable,
):
    if variance_operator.hilbert != vstate.hilbert:
        raise TypeError("Hilbert spaces should match")

    local_kernel = get_local_kernel(vstate, variance_operator.operator)
    local_kernel2 = get_local_kernel(vstate, variance_operator.operator_squared)

    sigma, args = get_local_kernel_arguments(vstate, variance_operator.operator)
    sigma, args2 = get_local_kernel_arguments(
        vstate, variance_operator.operator_squared
    )

    return expect_and_grad_inner_mc(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        local_kernel,
        local_kernel2,
        sigma,
        args,
        args2,
        return_grad=True,
    )


@partial(
    jax.jit, static_argnames=("afun", "local_kernel", "local_kernel2", "return_grad")
)
def expect_and_grad_inner_mc(
    afun,
    params,
    model_state,
    local_kernel,
    local_kernel2,
    sigma,
    args,
    args2,
    return_grad,
):
    n_chains = sigma.shape[0]

    σ = sigma.reshape(-1, sigma.shape[-1])

    def expect_kernel_var(params):
        log_pdf = lambda params, σ: 2 * afun({"params": params, **model_state}, σ).real

        def local_kernel_var(params, σ):
            W = {"params": params, **model_state}

            O_mean = nkjax.expect(
                log_pdf,
                lambda params, σ: local_kernel(
                    afun, {"params": params, **model_state}, σ, args
                ),
                params,
                σ,
                n_chains=n_chains,
            )[0]

            O2_loc = local_kernel2(afun, W, σ, args2)

            return (O2_loc - O_mean**2).real

        return nkjax.expect(log_pdf, local_kernel_var, params, σ, n_chains=n_chains)

    if not return_grad:
        var, var_stats = expect_kernel_var(params)
        return var_stats

    var, var_vjp_fun, var_stats = nkjax.vjp(
        expect_kernel_var, params, has_aux=True, conjugate=True
    )

    var_grad = var_vjp_fun(jnp.ones_like(var))[0]
    var_grad = jax.tree_util.tree_map(lambda x: mpi.mpi_sum_jax(x)[0], var_grad)

    return var_stats, var_grad



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from jax import numpy as jnp

from netket.hilbert import Qubit
from netket.utils.dispatch import dispatch


@dispatch
def random_state(hilb: Qubit, key, batches: int, *, dtype):
    rs = jax.random.randint(key, shape=(batches, hilb.size), minval=0, maxval=2)
    return jnp.asarray(rs, dtype=dtype)


@dispatch
def flip_state_scalar(hilb: Qubit, key, x, i):
    return x.at[i].set(-x[i] + 1), x[i]



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial, wraps
from typing import TYPE_CHECKING

import jax
from jax import numpy as jnp
from jax.tree_util import register_pytree_node_class

from netket.graph import AbstractGraph
from netket.hilbert import AbstractHilbert
from netket.utils.numbers import StaticZero
from netket.utils.types import DType

from .._discrete_operator_jax import DiscreteJaxOperator

from .base import IsingBase

if TYPE_CHECKING:
    from .numba import Ising


@register_pytree_node_class
class IsingJax(IsingBase, DiscreteJaxOperator):
    """
    Jax-compatible version of :class:`netket.operator.Ising`.
    """

    @wraps(IsingBase.__init__)
    def __init__(
        self,
        hilbert: AbstractHilbert,
        graph: AbstractGraph,
        h: float,
        J: float = 1.0,
        dtype: DType | None = None,
    ):
        if len(hilbert.local_states) != 2:
            raise ValueError(
                "IsingJax only supports Hamiltonians with two local states"
            )

        if not isinstance(h, jax.Array) and (h == 0 or h is None):
            h = StaticZero()

        J = jnp.array(J, dtype=dtype)
        if not isinstance(h, StaticZero):
            h = jnp.array(h, dtype=dtype)

        super().__init__(hilbert, graph=graph, h=h, J=J, dtype=dtype)

        self._edges = jnp.asarray(self.edges, dtype=jnp.int32)

    @jax.jit
    @wraps(IsingBase.n_conn)
    def n_conn(self, x):
        x_ids = self.hilbert.states_to_local_indices(x)
        return _ising_n_conn_jax(x_ids, self._edges, self.h, self.J)

    @jax.jit
    @wraps(IsingBase.get_conn_padded)
    def get_conn_padded(self, x):
        x_ids = self.hilbert.states_to_local_indices(x)
        xp_ids, mels = _ising_kernel_jax(x_ids, self._edges, self.h, self.J)
        xp = self.hilbert.local_indices_to_states(xp_ids, dtype=x.dtype)
        return xp, mels

    def to_numba_operator(self) -> "Ising":  # noqa: F821
        """
        Returns the standard (numba) version of this operator, which is an
        instance of {class}`nk.operator.Ising`.
        """

        from .numba import Ising

        return Ising(
            self.hilbert, graph=self.edges, h=self.h, J=self.J, dtype=self.dtype
        )

    def to_local_operator(self):
        # The hamiltonian
        ha = super().to_local_operator()

        return ha.to_jax_operator()

    def tree_flatten(self):
        data = (self.h, self.J, self.edges)
        metadata = {"hilbert": self.hilbert}
        return data, metadata

    @classmethod
    def tree_unflatten(cls, metadata, data):
        h, J, edges = data
        hi = metadata["hilbert"]

        res = cls(hi, h=1.0, graph=[(0, 0)])
        res._h = h
        res._J = J
        res._edges = edges
        return res


def _ising_mels_jax(x, edges, h, J):
    batch_dims = x.shape[:-1]
    if isinstance(h, StaticZero):
        max_conn_size = 1
    else:
        max_conn_size = x.shape[-1] + 1
    mels = jnp.zeros((*batch_dims, max_conn_size), dtype=J.dtype)

    same_spins = x[..., edges[:, 0]] == x[..., edges[:, 1]]
    mels = mels.at[..., 0].set(J * (2 * same_spins - 1).sum(axis=-1))
    if not isinstance(h, StaticZero):
        mels = mels.at[..., 1:].set(-h)
    return mels


def _ising_conn_states_jax(x, cond):
    was_state_0 = x == 0
    state_0 = jnp.asarray(0, dtype=x.dtype)
    state_1 = jnp.asarray(1, dtype=x.dtype)
    return jnp.where(cond ^ was_state_0, state_0, state_1)


@partial(jax.jit)
def _ising_kernel_jax(x, edges, h, J):
    hilb_size = x.shape[-1]
    batch_shape = x.shape[:-1]
    x = x.reshape((-1, hilb_size))

    mels = _ising_mels_jax(x, edges, h, J)
    mels = mels.reshape(batch_shape + mels.shape[1:])

    if isinstance(h, StaticZero):
        x_prime = jnp.expand_dims(x, axis=1)
    else:
        max_conn_size = hilb_size + 1
        flip = jnp.eye(max_conn_size, hilb_size, k=-1, dtype=bool)
        x_prime = _ising_conn_states_jax(x[..., None, :], flip)

    x_prime = x_prime.reshape(batch_shape + x_prime.shape[1:])

    return x_prime, mels


@jax.jit
def _ising_n_conn_jax(x, edges, h, J):
    n_conn_X = 0 if isinstance(h, StaticZero) else x.shape[-1]
    same_spins = x[..., edges[:, 0]] == x[..., edges[:, 1]]
    # TODO duplicated with _ising_mels_jax
    mels_ZZ = J * (2 * same_spins - 1).sum(axis=-1)
    n_conn_ZZ = jnp.asarray(mels_ZZ != 0, dtype=jnp.int32)
    return n_conn_X + n_conn_ZZ



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from collections.abc import Hashable, Iterable

from netket.utils.numbers import is_scalar
from netket.utils.types import DType, PyTree, Array

from netket.jax import canonicalize_dtypes
from netket.operator import ContinuousOperator
from netket.utils import struct, HashableArray

import jax.numpy as jnp


@struct.dataclass
class SumOperatorPyTree:
    """Internal class used to pass data from the operator to the jax kernel.

    This is used such that we can pass a PyTree containing some static data.
    We could avoid this if the operator itself was a pytree, but as this is not
    the case we need to pass as a separte object all fields that are used in
    the kernel.

    We could forego this, but then the kernel could not be marked as
    @staticmethod and we would recompile every time we construct a new operator,
    even if it is identical
    """

    ops: tuple[ContinuousOperator, ...] = struct.field(pytree_node=False)
    coeffs: Array
    op_data: tuple[PyTree, ...]


def _flatten_sumoperators(operators: Iterable[ContinuousOperator], coefficients: Array):
    """Flatten sumoperators inside of operators."""
    new_operators = []
    new_coeffs = []
    for op, c in zip(operators, coefficients):
        if isinstance(op, SumOperator):
            new_operators.extend(op.operators)
            new_coeffs.extend(c * op.coefficients)
        else:
            new_operators.append(op)
            new_coeffs.append(c)
    return new_operators, new_coeffs


class SumOperator(ContinuousOperator):
    r"""This class implements the action of the _expect_kernel()-method of
    ContinuousOperator for a sum of ContinuousOperator objects.
    """

    def __init__(
        self,
        *operators: tuple[ContinuousOperator, ...],
        coefficients: float | Iterable[float] = 1.0,
        dtype: DType | None = None,
    ):
        r"""
        Returns the action of a sum of local operators.
        Args:
            operators: A list of ContinuousOperator objects
            coefficients: A coefficient for each ContinuousOperator object
            dtype: Data type of the coefficients
        """
        hi_spaces = [op.hilbert for op in operators]
        if not all(hi == hi_spaces[0] for hi in hi_spaces):
            raise NotImplementedError(
                "Cannot add operators on different hilbert spaces"
            )

        if is_scalar(coefficients):
            coefficients = [coefficients for _ in operators]

        if len(operators) != len(coefficients):
            raise AssertionError("Each operator needs a coefficient")

        operators, coefficients = _flatten_sumoperators(operators, coefficients)

        dtype = canonicalize_dtypes(float, *operators, *coefficients, dtype=dtype)

        self._operators = tuple(operators)
        self._coefficients = jnp.asarray(coefficients, dtype=dtype)

        super().__init__(hi_spaces[0], self._coefficients.dtype)

        self._is_hermitian = all([op.is_hermitian for op in operators])
        self.__attrs = None

    @property
    def is_hermitian(self) -> bool:
        return self._is_hermitian

    @property
    def operators(self) -> tuple[ContinuousOperator, ...]:
        """The list of all operators in the terms of this sum. Every
        operator is summed with a corresponding coefficient
        """
        return self._operators

    @property
    def coefficients(self) -> Array:
        return self._coefficients

    @staticmethod
    def _expect_kernel(logpsi: Callable, params: PyTree, x: Array, data: PyTree | None):
        result = [
            data.coeffs[i] * op._expect_kernel(logpsi, params, x, op_data)
            for i, (op, op_data) in enumerate(zip(data.ops, data.op_data))
        ]

        return sum(result)

    def _pack_arguments(self) -> SumOperatorPyTree:
        return SumOperatorPyTree(
            self.operators,
            self.coefficients,
            tuple(op._pack_arguments() for op in self.operators),
        )

    @property
    def _attrs(self) -> tuple[Hashable, ...]:
        if self.__attrs is None:
            self.__attrs = (
                self.hilbert,
                self.operators,
                HashableArray(self.coefficients),
                self.dtype,
            )
        return self.__attrs

    def __repr__(self):
        return (
            f"SumOperator(operators={self.operators}, coefficients={self.coefficients})"
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce, partial
from collections.abc import Callable


import jax
import netket.jax as nkjax
from jax import numpy as jnp
from jax.tree_util import (
    register_pytree_node,
    tree_flatten,
    tree_unflatten,
    tree_map,
    tree_leaves,
)

from netket.utils.types import PyTree, Scalar
from netket.utils.numbers import is_scalar


def tree_ravel(pytree: PyTree) -> tuple[jnp.ndarray, Callable]:
    """Ravel (i.e. flatten) a pytree of arrays down to a 1D array.

    Args:
      pytree: a pytree to ravel

    Returns:
      A pair where the first element is a 1D array representing the flattened and
      concatenated leaf values, and the second element is a callable for
      unflattening a 1D vector of the same length back to a pytree of of the same
      structure as the input ``pytree``.
    """
    leaves, treedef = tree_flatten(pytree)
    flat, unravel_list = nkjax.vjp(_ravel_list, *leaves)
    unravel_pytree = lambda flat: tree_unflatten(treedef, unravel_list(flat))
    return flat, unravel_pytree


def _ravel_list(*lst):
    return jnp.concatenate([jnp.ravel(elt) for elt in lst]) if lst else jnp.array([])


def eval_shape(fun, *args, has_aux=False, **kwargs):
    """
    Returns the dtype of forward_fn(pars, v)
    """
    if has_aux:
        out, _ = jax.eval_shape(fun, *args, **kwargs)
    else:
        out = jax.eval_shape(fun, *args, **kwargs)
    return out


def tree_size(tree: PyTree) -> int:
    """
    Returns the sum of the size of all leaves in the tree.
    It's equivalent to the number of scalars in the pytree.
    """
    return sum(tree_leaves(tree_map(lambda x: x.size, tree)))


def tree_leaf_iscomplex(pars: PyTree) -> bool:
    """
    Returns true if at least one leaf in the tree has complex dtype.
    """
    return any(
        jax.tree_util.tree_leaves(jax.tree_util.tree_map(jnp.iscomplexobj, pars))
    )


def tree_leaf_isreal(pars: PyTree) -> bool:
    """
    Returns true if at least one leaf in the tree has real dtype.
    """
    return any(jax.tree_util.tree_leaves(jax.tree_util.tree_map(jnp.isrealobj, pars)))


def tree_ishomogeneous(pars: PyTree) -> bool:
    """
    Returns true if all leaves have real dtype or all leaves have complex dtype.
    """
    return not (tree_leaf_isreal(pars) and tree_leaf_iscomplex(pars))


@jax.jit
def tree_conj(t: PyTree) -> PyTree:
    r"""
    Conjugate all complex leaves. The real leaves are left untouched.
    Args:
        t: pytree
    """
    return jax.tree_util.tree_map(
        lambda x: jax.lax.conj(x) if jnp.iscomplexobj(x) else x, t
    )


@jax.jit
def tree_dot(a: PyTree, b: PyTree) -> Scalar:
    r"""
    Computes the dot product of two pytrees

    Args:
        a, b: pytrees with the same treedef

    Returns:
        A scalar equal the dot product of of the flattened arrays of a and b.
    """
    return jax.tree_util.tree_reduce(
        jax.numpy.add,
        jax.tree_util.tree_map(
            jax.numpy.sum, jax.tree_util.tree_map(jax.numpy.multiply, a, b)
        ),
    )


@partial(jax.jit, static_argnames="ord")
def tree_norm(a: PyTree, ord: int = 2) -> Scalar:
    r"""
    Compute the norm of a pytree, intended as a 1D vector of values.

    Equivalent to :code:`jnp.linalg.norm(nk.jax.tree_ravel(a)[0], ord)`.

    Args:
        a: A pytree, interpreted as a vector
        ord: Specify the vector L norm to be computed. Defaults to L=2.

    Returns:
        A scalar.
    """
    sum_norm = jax.tree_util.tree_reduce(
        jax.numpy.add,
        jax.tree_util.tree_map(
            jax.numpy.sum, jax.tree_util.tree_map(lambda x: jnp.abs(x) ** ord, a)
        ),
    )
    if ord == 2:
        return jnp.sqrt(sum_norm)
    elif ord == 1:
        return sum_norm
    else:
        return jnp.power(sum_norm, 1 / ord)


@jax.jit
def tree_cast(x: PyTree, target: PyTree) -> PyTree:
    r"""
    cast x the types of target

    Args:
        x: a pytree with arrays as leaves
        target: a pytree with the same treedef as x
                where only the dtypes of the leaves are accessed
    Returns:
        A pytree where each leaf of x is cast to the dtype of the corresponding leaf in target.
        The imaginary part of complex leaves which are cast to real is discarded.
    """
    # astype alone would also work, however that raises ComplexWarning when casting complex to real
    # therefore the real is taken first where needed
    return jax.tree_util.tree_map(
        lambda x, target: (x if jnp.iscomplexobj(target) else x.real).astype(
            target.dtype
        ),
        x,
        target,
    )


@jax.jit
def tree_axpy(a: Scalar | PyTree, x: PyTree, y: PyTree) -> PyTree:
    r"""
    Compute a * x + y

    Args:
      a: scalar or pytree
      x, y: pytrees with the same treedef
    Returns:
        The sum of the respective leaves of the two pytrees x and y
        where the leaves of x are first scaled with a.
    """
    ax = tree_ax(a, x)
    return jax.tree_util.tree_map(lambda ax_, y_: ax_ + y_, ax, y)


@jax.jit
def tree_ax(a: Scalar | PyTree, x: PyTree) -> PyTree:
    r"""
    Compute a * x , where a is a scalar or pytree.

    Args:
      a: scalar
      x: pytree
    Returns:
        The pytree x scaled by a
    """
    if is_scalar(a):
        return jax.tree_util.tree_map(lambda x_: a * x_, x)
    else:
        return jax.tree_util.tree_map(lambda a_, x_: a_ * x_, a, x)


class RealImagTuple(tuple):
    """
    A special kind of tuple which marks complex parameters which were split.
    Behaves like a regular tuple.
    """

    @property
    def real(self):
        return self[0]

    @property
    def imag(self):
        return self[1]


register_pytree_node(
    RealImagTuple,
    lambda xs: (xs, None),
    lambda _, xs: RealImagTuple(xs),
)


def _tree_to_real(x):
    if tree_leaf_iscomplex(x):
        # TODO find a way to make it a nop?
        # return jax.vmap(lambda y: jnp.array((y.real, y.imag)))(x)
        r = jax.tree_util.tree_map(lambda x: x.real if jnp.iscomplexobj(x) else x, x)
        i = jax.tree_util.tree_map(lambda x: x.imag if jnp.iscomplexobj(x) else None, x)
        return RealImagTuple((r, i))
    else:
        return x


def _tree_to_real_inverse(x):
    if isinstance(x, RealImagTuple):
        # not using jax.lax.complex because it would convert scalars to arrays
        return jax.tree_util.tree_map(
            lambda re, im: re + 1j * im if im is not None else re, *x
        )
    else:
        return x


def tree_to_real(pytree: PyTree) -> tuple[PyTree, Callable]:
    """Replace all complex leaves of a pytree with a RealImagTuple of 2 real leaves.

    Args:
      pytree: a pytree to convert to real

    Returns:
      A pair where the first element is the converted real pytree,
      and the second element is a callable for converting back a real pytree
      to a complex pytree of of the same structure as the input pytree.
    """
    return _tree_to_real(pytree), _tree_to_real_inverse


def compose(*funcs):
    """
    function composition

    compose(f,g,h)(x) is equivalent to f(g(h(x)))
    """

    def _compose(f, g):
        return lambda *args, **kwargs: f(g(*args, **kwargs))

    return reduce(_compose, funcs)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import functools
import inspect

from textwrap import dedent


def deprecated(reason=None, func_name=None):
    r"""
    This is a decorator which can be used to mark functions as deprecated. It
    will result in a warning being emitted when the function is used.
    """

    def decorator(func):
        object_type = "class" if inspect.isclass(func) else "function"
        fname = func_name or func.__name__
        message = f"Call to deprecated {object_type} {fname!r}"
        if reason is not None:
            message += f" ({dedent(reason)})"

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(message, category=FutureWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def warn_deprecation(message):
    r"""
    This is a function that sends a deprecation warning to the user about a
    function that will is now deprecated and will be removed in a future
    major release.

    :param message: A mandatory message documenting the deprecation.
    """
    warnings.warn(dedent(message), category=FutureWarning, stacklevel=2)


def deprecated_new_name(func_name, reason=""):
    def deprecated_decorator(func):
        @functools.wraps(func)
        def deprecated_func(*args, **kwargs):
            warnings.warn(
                dedent(
                    f"""

    {func.__name__} has been renamed to {func_name}. The old name is
    now deprecated and will be removed in the next minor version.

    Please update your code by chaing occurences of `{func.__name__}` with
    `{func_name}`.

    {dedent(reason)}

                    """
                ),
                category=FutureWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator


_dep_msg = """

**DEPRECATION_WARNING:**
    The `dtype` argument to neural-network layers and models is deprecated
    throughout NetKet to maintain consistency with new releases of flax.
    Please use `param_dtype` instead.

    This warning will become an error in a future version of NetKet.

"""


def _dtype_deprecated(self):
    warn_deprecation(_dep_msg)
    return self.param_dtype


def deprecation_getattr(module, deprecations):
    def getattr(name):
        if name in deprecations:
            message, fn = deprecations[name]
            if fn is None:  # Is the deprecation accelerated?
                raise AttributeError(message)
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            return fn
        raise AttributeError(f"module {module!r} has no attribute {name!r}")

    return getattr



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from netket.utils.types import DType as _DType
from netket.hilbert.abstract_hilbert import AbstractHilbert as _AbstractHilbert
from netket.experimental.operator import FermionOperator2nd as _FermionOperator2nd


def destroy(
    hilbert: _AbstractHilbert,
    site: int,
    sz: int | None = None,
    cutoff: float = 1e-10,
    dtype: _DType | None = None,
):
    """
    Builds the fermion destruction operator :math:`\\hat{a}` acting
    on the `site`-th of the Hilbert space `hilbert`.

    Args:
        hilbert: The hilbert space.
        site: the site on which this operator acts.
        sz: spin projection quantum number. This is the eigenvalue of
            the corresponding spin-Z Pauli operator (e.g. `sz = ±1` for
            a spin-1/2, `sz ∈ [-2, -1, 1, 2]` for a spin-3/2 and
            in general `sz ∈ [-2S, -2S + 2, ... 2S-2, 2S]` for
            a spin-S )
        dtype: The datatype to use for the matrix elements.

    Returns:
        The resulting FermionOperator2nd
    """
    idx = _get_index(hilbert, site, sz)
    return _FermionOperator2nd(hilbert, (((idx, 0),),), dtype=dtype)


def create(
    hilbert: _AbstractHilbert,
    site: int,
    sz: int | None = None,
    cutoff: float = 1e-10,
    dtype: _DType | None = None,
):
    """
    Builds the fermion creation operator :math:`\\hat{a}^\\dagger` acting
    on the `site`-th of the Hilbert space `hilbert`.

    Args:
        hilbert: The hilbert space
        site: the site on which this operator acts
        sz: spin projection quantum number. This is the eigenvalue of
            the corresponding spin-Z Pauli operator (e.g. `sz = ±1` for
            a spin-1/2, `sz ∈ [-2, -1, 1, 2]` for a spin-3/2 and
            in general `sz ∈ [-2S, -2S + 2, ... 2S-2, 2S]` for
            a spin-S )
        dtype: The datatype to use for the matrix elements.

    Returns:
        The resulting FermionOperator2nd
    """
    idx = _get_index(hilbert, site, sz)
    return _FermionOperator2nd(hilbert, (((idx, 1),),), dtype=dtype)


def number(
    hilbert: _AbstractHilbert,
    site: int,
    sz: int | None = None,
    cutoff: float = 1e-10,
    dtype: _DType | None = None,
):
    """
    Builds the number operator :math:`\\hat{a}^\\dagger\\hat{a}`  acting on the
    `site`-th of the Hilbert space `hilbert`.

    Args:
        hilbert: The hilbert space
        site: the site on which this operator acts
        site: the site on which this operator acts
        sz: spin projection quantum number. This is the eigenvalue of
            the corresponding spin-Z Pauli operator (e.g. `sz = ±1` for
            a spin-1/2, `sz ∈ [-2, -1, 1, 2]` for a spin-3/2 and
            in general `sz ∈ [-2S, -2S + 2, ... 2S-2, 2S]` for
            a spin-S )
        dtype: The datatype to use for the matrix elements.

    Returns:
        The resulting FermionOperator2nd
    """
    idx = _get_index(hilbert, site, sz)
    return _FermionOperator2nd(
        hilbert,
        (
            (
                (idx, 1),
                (idx, 0),
            ),
        ),
        dtype=dtype,
    )


def _get_index(hilbert: _AbstractHilbert, site: int, sz: int | None = None):
    """go from (site, spin_projection) indices to index in the (tensor) hilbert space"""
    if sz is None:
        if hasattr(hilbert, "spin") and hilbert.spin is not None:
            raise ValueError(
                "hilbert spaces with spin property require to specify the sz value to get the position in hilbert space"
            )
        return site
    elif not hasattr(hilbert, "spin"):
        raise ValueError("cannot specify sz for hilbert without spin property")
    elif hasattr(hilbert, "_get_index"):  # keep it general
        idx = hilbert._get_index(site, sz)
        if idx >= hilbert.size:
            raise IndexError(
                "requested site and sz combination is not present in the hilbert space"
            )
        return idx
    else:
        raise NotImplementedError(
            f"no method _get_index available for hilbert space {hilbert} that allows to find the position in hilbert space based on a spin projection value sz"
        )


def identity(
    hilbert: _AbstractHilbert, cutoff: float = 1e-10, dtype: _DType | None = None
):
    """
    Builds the :math:`\\mathbb{I}` identity operator.

    Args:
        hilbert: The hilbert space.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    return _FermionOperator2nd(hilbert, constant=1, dtype=dtype, cutoff=cutoff)


def zero(hilbert: _AbstractHilbert, cutoff: float = 1e-10, dtype: _DType | None = None):
    """
    Builds the :math:`0` operator, which has no connected components.

    Why we provide this is a mistery, as you could just multiply by 0.

    Args:
        hilbert: The hilbert space.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`."""
    return _FermionOperator2nd(hilbert, constant=0, dtype=dtype, cutoff=cutoff)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from typing import TYPE_CHECKING

import numpy as np

import jax
import jax.numpy as jnp
from jax.nn.initializers import normal

from flax import core as fcore
from flax.core.scope import CollectionFilter
from plum import Callable  # noqa: F401

from netket.hilbert.discrete_hilbert import DiscreteHilbert
import netket.jax as nkjax
from netket.operator import AbstractOperator
from netket.hilbert import AbstractHilbert
from netket.stats import Stats
from netket.utils.types import PyTree, PRNGKeyT, NNInitFunc
from netket.utils.dispatch import dispatch
from netket.utils.optional_deps import import_optional_dependency

if TYPE_CHECKING:
    from netket.optimizer.linear_operator import LinearOperator


QGTConstructor = Callable[["VariationalState"], "LinearOperator"]


class VariationalState(abc.ABC):
    """Abstract class for variational states representing either pure states
    or mixed quantum states.

    A variational state is a quantum state depending on a set of
    parameters, and that supports operations such
    as computing quantum expectation values and their gradients.

    A Variational stat can be serialized using flax's msgpack machinery.
    See `their docs <https://flax.readthedocs.io/en/latest/flax.serialization.html>`_.

    """

    mutable: CollectionFilter

    def __init__(self, hilbert: AbstractHilbert):
        """Initialize the Abstract base class of a Variational State defined
        on an hilbert space.

        Args:
            hilbert: The hilbert space upon which this state is defined.

        """
        self._hilbert = hilbert  # type: AbstractHilbert

        self._model_state = {}  # type: PyTree
        self._parameters = {}  # type: PyTree

    @property
    def hilbert(self) -> AbstractHilbert:
        r"""The descriptor of the Hilbert space
        on which this variational state is defined.
        """
        return self._hilbert

    @property
    def parameters(self) -> PyTree:
        r"""The pytree of the parameters of the model."""
        return fcore.copy(self._parameters, {})

    @parameters.setter
    def parameters(self, pars: PyTree):
        self._parameters = pars
        self.reset()

    @property
    def n_parameters(self) -> int:
        r"""The total number of parameters in the model."""
        return nkjax.tree_size(self.parameters)

    @property
    def model_state(self) -> PyTree:
        r"""The optional PyTree with the mutable state of the model, which
        is not optimized."""
        return fcore.copy(self._model_state, {})

    @model_state.setter
    def model_state(self, state: PyTree):
        self._model_state = state
        self.reset()

    @property
    def variables(self) -> PyTree:
        r"""The PyTree containing the parameters and state of the model,
        used when evaluating it.
        """
        return {"params": self.parameters, **self.model_state}

    @variables.setter
    def variables(self, var: PyTree):
        self.model_state, self.parameters = fcore.pop(var, "params")

    def init_parameters(
        self, init_fun: NNInitFunc | None = None, *, seed: PRNGKeyT | None = None
    ):
        r"""
        Re-initializes all the parameters with the provided initialization function,
        defaulting to the normal distribution of standard deviation 0.01.

        .. warning::

            The init function will not change the dtype of the parameters, which is
            determined by the model. DO NOT SPECIFY IT INSIDE THE INIT FUNCTION

        Args:
            init_fun: a jax initializer such as :func:`jax.nn.initializers.normal`.
                Must be a Callable taking 3 inputs, the jax PRNG key, the shape and the
                dtype, and outputting an array with the valid dtype and shape. If left
                unspecified, defaults to :code:`jax.nn.initializers.normal(stddev=0.01)`
            seed: Optional seed to be used. The seed is synced across all MPI processes.
                If unspecified, uses a random seed.
        """
        if init_fun is None:
            init_fun = normal(stddev=0.01)

        rng = nkjax.PRNGSeq(nkjax.PRNGKey(seed))

        def new_pars(par):
            return jnp.asarray(
                init_fun(rng.take(1)[0], shape=par.shape, dtype=par.dtype),  # type: ignore
                dtype=par.dtype,
            )

        self.parameters = jax.tree_util.tree_map(new_pars, self.parameters)

    def reset(self):
        r"""Resets the internal cache of th variational state.
        Called automatically when the parameters/state is updated.
        """

    def expect(self, O: AbstractOperator) -> Stats:
        r"""Estimates the quantum expectation value for a given operator
        :math:`O` or generic observable.
        In the case of a pure state :math:`\psi` and an operator, this is
        :math:`\langle O\rangle= \langle \Psi|O|\Psi\rangle/\langle\Psi|\Psi\rangle`
        otherwise for a  mixed state :math:`\rho`, this is
        :math:`\langle O\rangle= \textrm{Tr}[\rho \hat{O}]/\textrm{Tr}[\rho]`.

        Args:
            O: the operator or observable for which to compute the expectation
                value.

        Returns:
            An estimation of the quantum expectation value
            :math:`\langle O\rangle`.
        """
        return expect(self, O)

    def grad(
        self,
        Ô,
        *,
        use_covariance: bool | None = None,
        mutable: CollectionFilter | None = None,
    ) -> PyTree:
        r"""Estimates the gradient of the quantum expectation value of a given operator O.

        Args:
            op (netket.operator.AbstractOperator): the operator O.
            is_hermitian: optional override for whether to use or not the hermitian logic. By default
                it's automatically detected.

        Returns:
            array: An estimation of the average gradient of the quantum expectation value <O>.
        """
        r = self.expect_and_grad(Ô, use_covariance=use_covariance, mutable=mutable)
        return r[1]

    def expect_and_grad(
        self,
        O: AbstractOperator,
        *,
        mutable: CollectionFilter | None = None,
        **kwargs,
    ) -> tuple[Stats, PyTree]:
        r"""Estimates the quantum expectation value and its gradient
        for a given operator :math:`O`.

        Args:
            O: The operator :math:`O` for which expectation value and
                gradient are computed.
            mutable: Can be bool, str, or list. Specifies which collections in the
                     model_state should be treated as  mutable: bool: all/no collections
                     are mutable. str: The name of a single mutable  collection. list: A
                     list of names of mutable collections. This is used to mutate the state
                     of the model while you train it (for example to implement BatchNorm. Consult
                     `Flax's Module.apply documentation <https://flax.readthedocs.io/en/latest/_modules/flax/linen/module.html#Module.apply>`_
                     for a more in-depth explanation).
            use_covariance: whether to use the covariance formula, usually reserved for
                hermitian operators,
                :math:`\textrm{Cov}[\partial\log\psi, O_{\textrm{loc}}\rangle]`

        Returns:
            An estimate of the quantum expectation value <O>.
            An estimate of the gradient of the quantum expectation value <O>.
        """
        if mutable is None:
            mutable = self.mutable

        return expect_and_grad(self, O, mutable=mutable, **kwargs)

    def expect_and_forces(
        self,
        O: AbstractOperator,
        *,
        mutable: CollectionFilter | None = None,
    ) -> tuple[Stats, PyTree]:
        r"""Estimates the quantum expectation value and the corresponding force
        vector for a given operator O.

        The force vector :math:`F_j` is defined as the covariance of log-derivative
        of the trial wave function and the local estimators of the operator. For complex
        holomorphic states, this is equivalent to the expectation gradient
        :math:`\frac{\partial\langle O\rangle}{\partial(\theta_j)^\star} = F_j`. For real-parameter states,
        the gradient is given by :math:`\frac{\partial\partial_j\langle O\rangle}{\partial\partial_j\theta_j} = 2 \textrm{Re}[F_j]`.

        Args:
            O: The operator O for which expectation value and force are computed.
            mutable: Can be bool, str, or list. Specifies which collections in
                the model_state should be treated as  mutable: bool: all/no
                collections are mutable. str: The name of a single mutable
                collection. list: A list of names of mutable collections. This is
                used to mutate the state of the model while you train it (for
                example to implement BatchNorm. Consult
                `Flax's Module.apply documentation <https://flax.readthedocs.io/en/latest/_modules/flax/linen/module.html#Module.apply>`_
                for a more in-depth explanation).

        Returns:
            An estimate of the quantum expectation value <O>.
            An estimate of the force vector
            :math:`F_j = \textrm{Cov}[\partial_j\log\psi, O_{\textrm{loc}}]`.
        """
        if mutable is None:
            mutable = self.mutable

        return expect_and_forces(self, O, mutable=mutable)

    # @abc.abstractmethod
    def quantum_geometric_tensor(
        self, qgt_T: QGTConstructor | None
    ) -> "LinearOperator":
        r"""Computes an estimate of the quantum geometric tensor G_ij.

        This function returns a linear operator that can be used to apply G_ij to a
        given vector or can be converted to a full matrix.

        Args:
            qgt_T: the optional type of the quantum geometric tensor. By default it
                is automatically selected.

        Returns:
            nk.optimizer.LinearOperator: A linear operator representing the quantum
                geometric tensor.
        """
        raise NotImplementedError  # pragma: no cover

    def to_array(self, normalize: bool = True) -> jnp.ndarray:
        """
        Returns the dense-vector representation of this state.

        Args:
            normalize: If True, the vector is normalized to have L2-norm 1.

        Returns:
            An exponentially large vector representing the state in the computational
            basis.
        """
        return NotImplemented  # pragma: no cover

    def to_qobj(self):  # -> "qutip.Qobj"
        r"""Convert the variational state to a qutip's ket Qobj.

        Returns:
            A :class:`qutip.Qobj` object.
        """
        if not isinstance(self.hilbert, DiscreteHilbert):
            raise TypeError(
                "Only states defined on discrete hilbert spaces can be "
                "converted to the dense representation required by QuTiP."
            )

        qutip = import_optional_dependency("qutip", descr="to_qobj")

        q_dims = [list(self.hilbert.shape), [1 for i in range(self.hilbert.size)]]
        return qutip.Qobj(np.asarray(self.to_array()), dims=q_dims)


class VariationalMixedState(VariationalState):
    __module__ = "netket.vqs"

    def __init__(self, hilbert, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._hilbert_physical = hilbert

    @property
    def hilbert_physical(self) -> AbstractHilbert:
        return self._hilbert_physical

    def to_matrix(self, normalize: bool = True) -> jax.Array:
        """
        Returns the dense-matrix representation of this operator.

        Args:
            normalize: If True, the matrix is normalized to have trace 1.

        Returns:
            An exponentially large matrix representing the state in the computational
            basis.
        """
        return NotImplemented  # pragma: no cover

    def to_qobj(self):  # -> "qutip.Qobj"
        r"""Convert this mixed state to a qutip density matrix Qobj.

        Returns:
            A :class:`qutip.Qobj` object.
        """
        qutip = import_optional_dependency("qutip", descr="to_qobj")

        hilbert: DiscreteHilbert = self.hilbert_physical  # type: ignore
        q_dims = [list(hilbert.shape), list(hilbert.shape)]
        return qutip.Qobj(np.asarray(self.to_matrix()), dims=q_dims)


@dispatch.abstract
def expect(vstate: VariationalState, operator: AbstractOperator):
    """
    Computes the expectation value of the given operator over the
    variational state.

    Additional Information:
        To implement `vstate.expect` for a custom operator, implement
        the multiple-dispatch (plum-dispatc) based method according

        .. code:

            @nk.vqs.expect.register
            expect(vstate : VStateType operator: OperatorType):
                return ...

    Args:
        vstate: The VariationalState
        operator: The Operator or SuperOperator.

    Returns:
        The expectation value wrapped in a `Stats` object.
    """


# default dispatch where use_covariance is not specified
# Give it an higher precedence so this is always executed first, no matter what, if there
# is a dispatch ambiguity.
# This is not needed, but makes the dispatch logic work fine even if the users write weak
# signatures (eg: if an users defines `expect_grad(vs: MCState, op: MyOperator, use_cov: Any)`
# instead of `expect_grad(vs: MCState, op: MyOperator, use_cov: bool)`
# there would be a resolution error because the signature defined by the user is stricter
# for some arguments, but the one below here is stricter for `use_covariance` which is
# set to bool. Since this signature below, in the worst case, does nothing, this ensures
# that `expect_and_grad` is more user-friendly.
@dispatch.abstract
def expect_and_grad(
    vstate: VariationalState,
    operator: AbstractOperator,
    *args,
    mutable: CollectionFilter,
    **kwargs,
):
    r"""Estimates the quantum expectation value and its gradient for a given operator O.

    See `VariationalState.expect_and_grad` docstring for more information.

    Additional Information:
        To implement `vstate.expect` for a custom operator, implement
        the multiple-dispatch (plum-dispatc) based method according to the signature below.

        .. code:

            @nk.vqs.expect_and_grad.register
            expect_and_grad(vstate : VStateType, operator: OperatorType,
                            use_covariance : bool/Literal[True]/Literal[False], * mutable)
                return ...
    """


@dispatch.abstract
def expect_and_forces(
    vstate: VariationalState,
    operator: AbstractOperator,
    *args,
    mutable: CollectionFilter,
    **kwargs,
):
    r"""Estimates the quantum expectation value and corresponding force vector for a given operator O.

    See `VariationalState.expect_and_forces` docstring for more information.

    Additional Information:
        To implement `vstate.expect` for a custom operator, implement
        the multiple-dispatch (plum-dispatc) based method according to the signature below.

        .. code:

            @nk.vqs.expect_and_forces.register
            expect_and_forces(vstate : VStateType, operator: OperatorType,
                              use_covariance : bool/Literal[True]/Literal[False], * mutable)
                return ...
    """



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from flax import linen as nn
from flax.linen.dtypes import promote_dtype
from jax import numpy as jnp

from netket.utils import HashableArray

from .cells import RNNCell
from .ordering import check_reorder_idx


class RNNLayer(nn.Module):
    """Recurrent neural network layer that maps inputs at N sites to outputs at N sites."""

    cell: RNNCell
    """cell to update the hidden memory at each site, such as LSTM or GRU."""
    exclusive: bool
    """True if an output element does not depend on the input element at the same index."""
    reorder_idx: HashableArray | None = None
    """indices to transform the inputs from unordered to ordered.
    See :meth:`netket.models.AbstractARNN.reorder` for details.
    """
    inv_reorder_idx: HashableArray | None = None
    """indices to transform the inputs from ordered to unordered.
    See :meth:`netket.models.AbstractARNN.reorder` for details.
    """
    prev_neighbors: HashableArray | None = None
    """previous neighbors of each site."""
    unroll: int = 1
    """how many steps to unroll in the recurrent loop. Trades compile time for
    faster runtime when networks are small."""

    def __post_init__(self):
        super().__post_init__()
        check_reorder_idx(self.reorder_idx, self.inv_reorder_idx, self.prev_neighbors)

    def _extract_inputs_i(self, inputs, k, index, prev_index):
        assert inputs.ndim == 3

        # Masking for 'exclusive' behaviour of first layer
        # TODO: Use 0 in masked sites or a value from computational basis?
        if self.exclusive:
            # Get the inputs at the previous site in the autoregressive order,
            # or zeros for the first site
            inputs_i = inputs[:, prev_index, :]
            inputs_i = jnp.where(k == 0, 0, inputs_i)
        else:
            # Get the inputs at the current site
            inputs_i = inputs[:, index, :]
        return inputs_i

    def _extract_hidden(self, outputs, index, prev_neighbors):
        assert outputs.ndim == 3

        if self.reorder_idx is None:
            # Get the hidden memory at the previous site,
            # or zeros for the first site
            hidden = outputs[:, index - 1, :]
            hidden = jnp.expand_dims(hidden, axis=-1)
        else:
            # Get the hidden memories at the previous neighbors
            prev_neighbors_i = prev_neighbors[index]
            hidden = outputs[:, prev_neighbors_i, :]
            # mask out inexistant previous neighbords
            hidden = jnp.where(prev_neighbors_i[None, :, None] == -1, 0, hidden)
        return hidden

    def __call__(self, inputs):
        """
        Applies the RNN cell to a batch of input sequences.

        Args:
          inputs: input data with dimensions (batch, n_sites, features).

        Returns:
          The output sequences.
        """
        if inputs.ndim != 3:
            raise ValueError(
                "Requires 3 dimensions of (batch, n_sites, features). "
                "If you have no features, set it to 1."
            )

        batch_size, N, _ = inputs.shape
        inputs = promote_dtype(inputs, dtype=self.cell.param_dtype)[0]

        if self.reorder_idx is None:
            reorder_idx = None
            prev_neighbors = None
        else:
            reorder_idx = jnp.asarray(self.reorder_idx)
            prev_neighbors = jnp.asarray(self.prev_neighbors)

        def scan_func(rnn_cell, carry, k):
            cell_mem, outputs = carry
            if self.reorder_idx is None:
                index = k
                prev_index = k - 1
            else:
                index = reorder_idx[k]
                prev_index = reorder_idx[k - 1]

            inputs_i = self._extract_inputs_i(inputs, k, index, prev_index)
            hidden = self._extract_hidden(outputs, index, prev_neighbors)
            cell_mem, hidden = rnn_cell(inputs_i, cell_mem, hidden)
            outputs = outputs.at[:, index, :].set(hidden)
            return (cell_mem, outputs), None

        scan = nn.scan(
            scan_func,
            variable_broadcast="params",
            split_rngs={"params": False},
            unroll=self.unroll,
        )

        cell_mem = jnp.zeros((batch_size, self.cell.features), dtype=inputs.dtype)
        outputs = jnp.zeros((batch_size, N, self.cell.features), dtype=inputs.dtype)
        (_, outputs), _ = scan(self.cell, (cell_mem, outputs), jnp.arange(N))
        return outputs



# Copyright 2021 The NetKet Authors - All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Union
from collections.abc import Callable
from collections.abc import Sequence

import optax as _optax
import jax as _jax
import numpy as _np

PRNGKeyT = Any
SeedT = Union[int, PRNGKeyT]

Shape = Sequence[int]
DType = Any  # this could be a real type?

Array = Union[_np.ndarray, _jax.Array]
JaxArray = _jax.Array

ArrayLike = Any  # Objects that are valid inputs to (np|jnp).asarray.

NNInitFunc = Callable[[PRNGKeyT, Shape, DType], _jax.Array]

PyTree = Any

Scalar = Any

ScalarOrSchedule = Union[Scalar, _optax.Schedule]

Optimizer = Any



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from functools import partial


class HashablePartial(partial):
    """
    A class behaving like functools.partial, but that retains it's hash
    if it's created with a lexically equivalent (the same) function and
    with the same partially applied arguments and keywords.

    It also stores the computed hash for faster hashing.
    """

    # TODO remove when dropping support for Python < 3.10
    def __new__(cls, func, *args, **keywords):
        # In Python 3.10+ if func is itself a functools.partial instance,
        # functools.partial.__new__ would merge the arguments of this HashablePartial
        # instance with the arguments of the func
        # Pre 3.10 this does not happen, so here we emulate this behaviour recursively
        # This is necessary since functools.partial objects do not have a __code__
        # property which we use for the hash
        # For python 3.10+ we still need to take care of merging with another HashablePartial
        while isinstance(
            func, partial if sys.version_info < (3, 10) else HashablePartial
        ):
            original_func = func
            func = original_func.func
            args = original_func.args + args
            keywords = {**original_func.keywords, **keywords}
        return super().__new__(cls, func, *args, **keywords)

    def __init__(self, *args, **kwargs):
        self._hash = None

    def __eq__(self, other):
        return (
            type(other) is HashablePartial
            and self.func.__code__ == other.func.__code__
            and self.args == other.args
            and self.keywords == other.keywords
        )

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(
                (self.func.__code__, self.args, frozenset(self.keywords.items()))
            )

        return self._hash

    def __repr__(self):
        return f"<hashable partial {self.func.__name__} with args={self.args} and kwargs={self.keywords}, hash={hash(self)}>"


# jax.tree_util.register_pytree_node(
#    HashablePartial,
#    lambda partial_: ((), (partial_.func, partial_.args, partial_.keywords)),
#    lambda args, _: StaticPartial(args[0], *args[1], **args[2]),
# )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from flax import linen as nn
from flax.linen.dtypes import promote_dtype

from jax import lax
from jax import numpy as jnp
from jax.nn.initializers import zeros

from netket.nn.masked_linear import (
    MaskedConv1D,
    MaskedConv2D,
    MaskedDense1D,
    _conv_dimension_numbers,
    default_kernel_init,
    wrap_kernel_init,
)
from netket.utils.types import Array, DType, NNInitFunc


class FastMaskedDense1D(nn.Module):
    """
    1D linear transformation module with mask for fast autoregressive NN.

    See :class:`netket.models.FastARNNSequential` for a brief explanation of fast autoregressive sampling.

    TODO: FastMaskedDense1D does not support JIT yet, because it involves slicing the cached inputs
    and the weights with a dynamic shape.
    """

    size: int
    """number of sites."""
    features: int
    """output feature density, should be the last dimension."""
    exclusive: bool
    """True if an output element does not depend on the input element at the same index."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see :class:`jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the weight matrix."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    @nn.compact
    def update_site(self, inputs: Array, index: int) -> Array:
        """
        Adds an input site into the cache, and applies the masked linear transformation to the cache.

        Args:
          inputs: an input site to be added into the cache with dimensions (batch, features).
          index: the index of the output site. The index of the input site should be `index - self.exclusive`.

        Returns:
          The output site with dimensions (batch, features).
        """
        if inputs.ndim == 1:
            is_single_input = True
            inputs = jnp.expand_dims(inputs, axis=0)
        else:
            is_single_input = False

        batch, in_features = inputs.shape
        size = self.size

        if self.use_bias:
            bias = self.param(
                "bias", self.bias_init, (size, self.features), self.param_dtype
            )
        else:
            bias = None

        # The construction of `mask` will be optimized to a constant by JIT
        mask = jnp.ones((size, size), dtype=self.param_dtype)
        mask = jnp.triu(mask, self.exclusive)
        mask = jnp.kron(
            mask, jnp.ones((in_features, self.features), dtype=self.param_dtype)
        )

        kernel = self.param(
            "kernel",
            wrap_kernel_init(self.kernel_init, mask),
            (size * in_features, size * self.features),
            self.param_dtype,
        )

        inputs, kernel, mask, bias = promote_dtype(
            inputs, kernel, mask, bias, dtype=None
        )

        # Number of input sites depended by the output site at the index
        size_i = index + 1

        # Initialize the cache with zeros, and the RNG key is None
        # `cache.dtype` must be the same as `inputs.dtype` (no promotion)
        _cache = self.variable(
            "cache", "inputs", zeros, None, (batch, size, in_features), inputs.dtype
        )

        if not self.is_initializing():
            # Add the input site into the cache
            # To write the cache, use `_cache.value` as the left value of the assignment
            _cache.value = jnp.where(
                index - self.exclusive >= 0,
                _cache.value.at[:, index - self.exclusive, :].set(inputs),
                _cache.value,
            )

        cache = _cache.value

        cache_i = cache[:, :size_i, :]
        cache_i = cache_i.reshape((batch, size_i * in_features))

        mask_i = mask.reshape((size, in_features, size, self.features))
        mask_i = mask_i[:size_i, :, index, :]
        mask_i = mask_i.reshape((size_i * in_features, self.features))

        kernel_i = kernel.reshape((size, in_features, size, self.features))
        kernel_i = kernel_i[:size_i, :, index, :]
        kernel_i = kernel_i.reshape((size_i * in_features, self.features))

        y_i = lax.dot(cache_i, mask_i * kernel_i, precision=self.precision)

        if self.use_bias:
            y_i = y_i + bias[index, :]  # type: ignore[assignment,index]

        assert y_i.shape[1] == self.features

        if is_single_input:
            y_i = y_i.squeeze(axis=0)

        return y_i

    def __call__(self, inputs: Array) -> Array:
        """
        Applies the masked linear transformation to all input sites.

        Args:
          inputs: input data with dimensions (batch, size, features).

        Returns:
          The transformed data.
        """
        return MaskedDense1D.__call__(self, inputs)  # type: ignore[arg-type]


class FastMaskedConv1D(nn.Module):
    """
    1D convolution module with mask for fast autoregressive NN.

    See :class:`netket.models.FastARNNSequential` for a brief explanation of fast autoregressive sampling.
    """

    features: int
    """number of convolution filters."""
    kernel_size: int
    """length of the convolutional kernel."""
    kernel_dilation: int
    """dilation factor of the convolution kernel."""
    exclusive: bool
    """True if an output element does not depend on the input element at the same index."""
    feature_group_count: int = 1
    """if specified, divides the input features into groups (default: 1)."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see :class:`jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the convolutional kernel."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    @nn.compact
    def update_site(self, inputs: Array, index: int) -> Array:
        """
        Adds an input site into the cache, and applies the masked convolution to the cache.

        Args:
          inputs: an input site to be added into the cache with dimensions (batch, features).
          index: the index of the output site. The index of the input site should be `index - self.exclusive`.

        Returns:
          The next output site with dimensions (batch, features).
        """
        kernel_size = self.kernel_size - self.exclusive
        dilation = self.kernel_dilation

        if inputs.ndim == 1:
            is_single_input = True
            inputs = jnp.expand_dims(inputs, axis=0)
        else:
            is_single_input = False

        batch, in_features = inputs.shape
        assert in_features % self.feature_group_count == 0
        cache_size = kernel_size * dilation - (not self.exclusive) * (dilation - 1)

        kernel_shape = (
            kernel_size,
            in_features // self.feature_group_count,
            self.features,
        )
        kernel = self.param("kernel", self.kernel_init, kernel_shape, self.param_dtype)

        if self.use_bias:
            bias = self.param(
                "bias", self.bias_init, (self.features,), self.param_dtype
            )
        else:
            bias = None

        inputs, kernel, bias = promote_dtype(inputs, kernel, bias, dtype=None)

        # Initialize the cache with zeros, and the RNG key is None
        # `cache.dtype` must be the same as `inputs.dtype` (no promotion)
        _cache = self.variable(
            "cache",
            "inputs",
            zeros,
            None,
            (batch, cache_size, in_features),
            inputs.dtype,
        )

        if not self.is_initializing():
            # Add the input site into the cache
            # To write the cache, use `_cache.value` as the left value of the assignment
            _cache.value = jnp.where(
                index - self.exclusive >= 0,
                jnp.concatenate(
                    [_cache.value[:, 1:, :], jnp.expand_dims(inputs, axis=1)], axis=1
                ),
                _cache.value,
            )

        cache = _cache.value

        if self.exclusive and dilation > 1:
            cache = cache[:, : -(dilation - 1), :]

        dimension_numbers = _conv_dimension_numbers(cache.shape)
        y_i = lax.conv_general_dilated(
            cache,
            kernel,  # type: ignore[arg-type]
            window_strides=(1,),
            padding="VALID",
            lhs_dilation=(1,),
            rhs_dilation=(dilation,),
            dimension_numbers=dimension_numbers,
            feature_group_count=self.feature_group_count,
            precision=self.precision,
        )

        if self.use_bias:
            y_i = y_i + bias

        y_i = y_i.squeeze(axis=1)

        if is_single_input:
            y_i = y_i.squeeze(axis=0)

        return y_i

    def __call__(self, inputs: Array) -> Array:
        """
        Applies the masked convolution to all input sites.

        Args:
          inputs: input data with dimensions (batch, size, features).

        Returns:
          The convolved data.
        """
        return MaskedConv1D.__call__(self, inputs)


class FastMaskedConv2D(nn.Module):
    """
    2D convolution module with mask for fast autoregressive NN.

    See :class:`netket.models.FastARNNSequential` for a brief explanation of fast autoregressive sampling.
    """

    L: int
    """edge length of the 2D lattice."""
    features: int
    """number of convolution filters."""
    kernel_size: tuple[int, int]
    """shape of the convolutional kernel `(h, w)`. Typically, :math:`h = w // 2 + 1`."""
    kernel_dilation: tuple[int, int]
    """a sequence of 2 integers, giving the dilation factor to
    apply in each spatial dimension of the convolution kernel."""
    exclusive: bool
    """True if an output element does not depend on the input element at the same index."""
    feature_group_count: int = 1
    """if specified, divides the input features into groups (default: 1)."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see :class:`jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the convolutional kernel."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    def setup(self):
        MaskedConv2D.setup(self)

    @nn.compact
    def update_site(self, inputs: Array, index: int) -> Array:
        """
        Adds an input site into the cache, and applies the masked convolution to the cache.

        Args:
          inputs: an input site to be added into the cache with dimensions (batch, features).
          index: the index of the output site. The index of the input site should be `index - self.exclusive`.

        Returns:
          The next output site with dimensions (batch, features).
        """
        L = self.L
        index_w = index % L

        kernel_h, kernel_w = self.kernel_size
        dilation_h, dilation_w = self.kernel_dilation
        ones = (1, 1)

        if inputs.ndim == 1:
            is_single_input = True
            inputs = jnp.expand_dims(inputs, axis=0)
        else:
            is_single_input = False

        batch, in_features = inputs.shape
        assert in_features % self.feature_group_count == 0
        recep_h = (kernel_h - 1) * dilation_h + 1
        recep_w = (kernel_w - 1) * dilation_w + 1

        if self.use_bias:
            bias = self.param(
                "bias", self.bias_init, (self.features,), self.param_dtype
            )
        else:
            bias = None

        kernel_shape = self.kernel_size + (
            in_features // self.feature_group_count,
            self.features,
        )
        kernel = self.param(
            "kernel",
            wrap_kernel_init(self.kernel_init, self.mask),
            kernel_shape,
            self.param_dtype,
        )

        inputs, kernel, bias = promote_dtype(inputs, kernel, bias, dtype=None)

        # Initialize the cache with zeros, and the RNG key is None
        # `cache.dtype` must be the same as `inputs.dtype` (no promotion)
        _cache = self.variable(
            "cache",
            "inputs",
            zeros,
            None,
            (batch, recep_h, L, in_features),
            inputs.dtype,
        )

        if not self.is_initializing():
            # Add the input site into the cache
            # To write the cache, use `_cache.value` as the left value of the assignment

            inputs = jnp.expand_dims(inputs, axis=(1, 2))

            # Index of the input site in the width direction
            index_w_in = (index - self.exclusive) % L

            def _add(cache):
                # return cache.at[:, -1, index_w_in, :].set(inputs)
                return lax.dynamic_update_slice(cache, inputs, (0, -1, index_w_in, 0))

            def _shift(cache):
                return jnp.pad(cache[:, 1:, :, :], ((0, 0), (0, 1), (0, 0), (0, 0)))

            cache_new_row = jnp.where(
                index_w_in == 0, _add(_shift(_cache.value)), _shift(_add(_cache.value))
            )

            cache_new = jnp.where(index_w == 0, cache_new_row, _add(_cache.value))

            _cache.value = jnp.where(
                index - self.exclusive >= 0, cache_new, _cache.value
            )

        cache = _cache.value

        # Zero padding
        cache = jnp.pad(
            cache,
            (
                (0, 0),
                (0, 0),
                (kernel_w // 2 * dilation_w, (kernel_w - 1) // 2 * dilation_w),
                (0, 0),
            ),
        )

        # cache = cache[:, :, index_w : index_w + recep_w, :]
        cache = lax.dynamic_slice(
            cache, (0, 0, index_w, 0), (batch, recep_h, recep_w, in_features)
        )

        dimension_numbers = _conv_dimension_numbers(cache.shape)
        y_i = lax.conv_general_dilated(
            cache,
            kernel,
            window_strides=ones,
            padding="VALID",
            lhs_dilation=ones,
            rhs_dilation=self.kernel_dilation,
            dimension_numbers=dimension_numbers,
            feature_group_count=self.feature_group_count,
            precision=self.precision,
        )

        if self.use_bias:
            y_i = y_i + bias

        y_i = y_i.squeeze(axis=(1, 2))

        if is_single_input:
            y_i = y_i.squeeze(axis=0)

        return y_i

    def __call__(self, inputs: Array) -> Array:
        """
        Applies the masked convolution to all input sites.

        Args:
          inputs: input data with dimensions (batch, width, height, features).

        Returns:
          The convolved data.
        """
        return MaskedConv2D.__call__(self, inputs)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
from .continuous_hilbert import ContinuousHilbert


class Particle(ContinuousHilbert):
    r"""Hilbert space derived from ContinuousHilbert defining N particles
    in continuous space with or without periodic boundary conditions."""

    def __init__(
        self,
        N: int | tuple[int, ...],
        L: tuple[float, ...] | None = None,
        pbc: bool | tuple[bool, ...] | None = None,
        *,
        D: int | None = None,
    ):
        """
        Constructs new ``Particles`` given specifications
        of the continuous space they are defined in.

        Args:
            N: Number of particles. If int all have the same spin. If Tuple the entry indicates how many particles
                there are with a certain spin-projection.
            L: Tuple indicating the maximum of the continuous quantum number(s) in the configurations. Each entry
                in the tuple corresponds to a different physical dimension.
                If `np.inf` is used an infinite box is considered and `pbc=False` is mandatory (because what are PBC
                if there are no boundaries?). If a finite value is given, a minimum value of zero is assumed for the
                quantum number(s).
                A particle in a 3D box of size L would take `(L,L,L)`. A rotor model would take e.g. `(2pi,)`.
            pbc: Tuple or bool indicating whether to use periodic boundary conditions in a given physical dimension.
                If tuple it must have the same length as domain. If bool the same value is used for all the dimensions
                defined in domain.
            D: (Optional) Number of dimensions. Can be specified instead of `L` and `pbc` in order to construct a
                `Particle` in a $D-$ dimensional infinite box. Equivalent to
                `Particle(N, L=(np.inf,) * D, pbc=False)`.
        """
        if D is None and L is None:
            raise ValueError("Must specify at least `L` or `D`.")
        elif D is not None:
            if L is not None:
                raise TypeError(
                    "Cannot specify at the same time `D` and `L`. If you want to use "
                    "an infinite box, just specify D, otherwise specify L."
                )
            L = (np.inf,) * D

        # Assume 1D if L is a scalar
        if not hasattr(L, "__len__"):
            L = (L,)

        if not hasattr(N, "__len__"):
            N = (N,)

        if pbc is None:
            if np.all(np.isinf(L)):
                pbc = False
            else:
                raise ValueError("`pbc` must be specified if `L` is finite.")

        if isinstance(pbc, bool):
            pbc = [pbc] * len(L)

        if np.any(np.logical_and(np.isinf(L), pbc)):
            raise ValueError(
                "Cannot combine periodic boundary conditions and infinite size along the same dimension."
            )

        self._N = sum(N)
        self._n_per_spin = N

        super().__init__(L, pbc)

    @property
    def size(self) -> int:
        return self._N * len(self._extent)

    @property
    def n_particles(self) -> int:
        r"""The number of particles"""
        return self._N

    @property
    def n_per_spin(self) -> tuple[int, ...]:
        r"""Gives the number of particles in a specific spin
        projection.

        The length of this tuple indicates the total spin whereas
        the position in the tuple indicates the spin projection.

        Example: (10,5,3) describes 18 particles of total spin 1
        where 10 of those have spin-projection -1, 5 have
        spin-projection 0 and 3 have spin-projection 1.
        """
        return self._n_per_spin

    @property
    def _attrs(self):
        return (self._N, self.extent, self.pbc)

    def __repr__(self):
        return f"ContinuousParticle(N={self.n_particles}, " f"d={len(self.extent)})"



# Copyright 2022-23 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers

import numpy as np

from scipy import sparse
from scipy.sparse import spmatrix

from netket.hilbert import AbstractHilbert, Fock
from netket.jax import canonicalize_dtypes
from netket.utils.types import DType, Array


def cast_operator_matrix_dtype(matrix: Array, dtype: DType):
    """
    Changes the dtype of a matrix, without changing the structural type of the object.

    This makes sure that if you pass sparse arrays to a LocalOperator, they remain
    sparse even if you change the dtype
    """
    # must copy
    # return np.asarray(matrix, dtype=dtype)
    return matrix.astype(dtype)


def _standardize_matrix_input_type(op):
    """
    Standardize the structural type of operators stored in LocalOperator.

    Eventually, we could also support spmatrices (but some work will be needed.)
    """
    if isinstance(op, list):
        return np.asarray(op)
    elif sparse.issparse(op):
        return op.tocoo()
    elif isinstance(op, np.matrix):
        # np.matrix does not respect the ndarray interface
        return np.asarray(op)
    else:
        return op


def _kron(A, B):
    if isinstance(A, spmatrix) and isinstance(B, spmatrix):
        res = sparse.kron(A, B, format="coo")
    elif isinstance(A, spmatrix) ^ isinstance(B, spmatrix):
        # XOR, meaning only one is sparse matrix
        # Then we maintain sparse format (Is this a good idea?)
        res = sparse.kron(A, B, format="coo")
    else:
        res = np.kron(np.asarray(A), np.asarray(B))
    return res


def _eye_like(size, *, dtype, like):
    if sparse.issparse(like):
        return sparse.eye(size, dtype=dtype)
    else:
        return np.eye(size, dtype=dtype)


def canonicalize_input(
    hilbert: AbstractHilbert, operators, acting_on, constant, *, dtype=None
):
    """
    Takes as inputs the inputs to the constructor of LocalOperator and canonicalizes
    them by ensuring the following holds:
     - acting_on is a list of list
     - acting_on[i] are sorted
     - operators is list of matrices
     - all dtypes match

    Args:
        hilbert: The hilbert space

    Returns:
        List of operators, acting ons and dtypes.
    """
    # check if passing a single operator or a list of operators
    if isinstance(acting_on, numbers.Number):
        acting_on = [acting_on]

    is_nested = any(hasattr(i, "__len__") for i in acting_on)
    if not is_nested:
        operators = [operators]
        acting_on = [acting_on]

    if all(len(aon) == 0 for aon in acting_on):
        operators = []
        acting_on = []
    else:
        if max(map(max, acting_on)) >= hilbert.size or min(map(min, acting_on)) < 0:
            raise ValueError("An operator acts on an invalid set of sites.")

        for i, aon in enumerate(acting_on):
            if len(aon) != len(set(aon)):
                raise ValueError(
                    f"The operator at index {i} acts on duplicated sites {aon}"
                )

    acting_on = [tuple(aon) for aon in acting_on]
    # operators = [np.asarray(operator) for operator in operators]
    operators = [_standardize_matrix_input_type(op) for op in operators]

    dtype = canonicalize_dtypes(float, *operators, constant, dtype=dtype)

    canonicalized_operators = []
    canonicalized_acting_on = []
    for operator, acting_on in zip(operators, acting_on):
        check_valid_opmatrix(hilbert, operator, acting_on)

        if operator.dtype is not dtype:
            operator = cast_operator_matrix_dtype(operator, dtype=dtype)

        # re-sort the operator
        operator, acting_on = _reorder_kronecker_product(hilbert, operator, acting_on)
        canonicalized_operators.append(operator)
        canonicalized_acting_on.append(acting_on)

    return canonicalized_operators, canonicalized_acting_on, dtype


def check_valid_opmatrix(hi, mat, acting_on):
    """ """
    expected_size = np.prod([hi.shape[aon] for aon in acting_on])

    if mat.shape != (expected_size, expected_size):
        raise ValueError(
            f"The matrix of the sub-operator acting on sites {acting_on} "
            f"must have shape {expected_size, expected_size}, "
            f"but it has shape {mat.shape}."
        )


# TODO: support sparse arrays without returning dense arrays
def _reorder_kronecker_product(hi, mat, acting_on) -> tuple[Array, tuple]:
    """
    Reorders the matrix resulting from a kronecker product of several
    operators in such a way to sort acting_on.

    A conceptual example is the following:
    if `mat = Â ⊗ B̂ ⊗ Ĉ` and `acting_on = [[2],[1],[3]`
    you will get `result = B̂ ⊗ Â ⊗ Ĉ, [[1], [2], [3]].

    However, essentially, A,B,C represent some operators acting on
    thei sub-space acting_on[1], [2] and [3] of the hilbert space.

    This function also handles any possible set of values in acting_on.

    The inner logic uses the Fock.all_states(), number_to_state and
    state_to_number to perform the re-ordering.
    """
    acting_on_sorted = np.sort(acting_on)
    if np.array_equal(acting_on_sorted, acting_on):
        return mat, acting_on

    # could write custom binary <-> int logic instead of using Fock...
    # Since i need to work with bit-strings (where instead of bits i
    # have integers, in order to support arbitrary size spaces) this
    # is exactly what hilbert.to_number() and vice versa do.

    # target ordering binary representation
    hi_subspace = Fock(hi.shape[acting_on_sorted[0]] - 1)
    for site in acting_on_sorted[1:]:
        hi_subspace = hi_subspace * Fock(hi.shape[site] - 1)

    hi_unsorted_subspace = Fock(hi.shape[acting_on[0]] - 1)
    for site in acting_on[1:]:
        hi_unsorted_subspace = hi_unsorted_subspace * Fock(hi.shape[site] - 1)

    # find how to map target ordering back to unordered
    acting_on_unsorted_ids = np.zeros(len(acting_on), dtype=np.intp)
    for i, site in enumerate(acting_on):
        acting_on_unsorted_ids[i] = np.argmax(site == acting_on_sorted)

    # now it is valid that
    # acting_on_sorted == acting_on[acting_on_unsorted_ids]

    # generate n-bit strings in the target ordering
    v = hi_subspace.all_states()

    # convert them to origin (unordered) ordering
    v_unsorted = v[:, acting_on_unsorted_ids]
    # convert the unordered bit-strings to numbers in the target space.
    n_unsorted = hi_unsorted_subspace.states_to_numbers(v_unsorted)

    # reorder the matrix
    if sparse.issparse(mat):
        # covert the indices from going from old -> new to new->old
        idxs = np.argsort(n_unsorted)

        # sparse matrices don't support indexing, and can be sorted more
        # quickly
        mat_sorted = mat.copy()
        mat_sorted.row[:] = idxs[mat.row]
        mat_sorted.col[:] = idxs[mat.col]
    else:
        mat_sorted = mat[n_unsorted, :][:, n_unsorted]

    return mat_sorted, tuple(acting_on_sorted)


# TODO: support sparse arrays without returning dense arrays
def _multiply_operators(
    hilbert, support_A: tuple, A: Array, support_B: tuple, B: Array, *, dtype
) -> tuple[tuple, Array]:
    """
    Returns the `Tuple[acting_on, Matrix]` representing the operator obtained by
    multiplying the two input operators A and B.
    """
    support_A = np.asarray(support_A)
    support_B = np.asarray(support_B)

    inters = np.intersect1d(support_A, support_B, return_indices=False)

    if support_A.size == support_B.size and np.array_equal(support_A, support_B):
        # TODO: Note: COO @ COO -> CSR, this may insert CSR matrices in our
        # format. Should we enforce COO?
        # If yes, the logic in the compilation can be simplified to always assume
        # COO.
        return tuple(support_A), A @ B
    elif inters.size == 0:
        # disjoint supports
        support = tuple(np.concatenate([support_A, support_B]))
        operator = _kron(A, B)
        operator, support = _reorder_kronecker_product(hilbert, operator, support)
        return tuple(support), operator
    else:
        _support_A = list(support_A)
        _support_B = list(support_B)
        _A = A.copy()
        _B = B.copy()

        # expand _act to match _act_i
        supp_B_min = min(support_B)
        for site in support_A:
            if site not in support_B:
                I = _eye_like(hilbert.shape[site], dtype=dtype, like=B)
                if site < supp_B_min:
                    _support_B = [site, *_support_B]
                    _B = _kron(I, _B)
                else:  # site > actmax
                    _support_B = [*_support_B, site]
                    _B = _kron(_B, I)

        supp_A_min = min(support_A)
        for site in support_B:
            if site not in support_A:
                I = _eye_like(hilbert.shape[site], dtype=dtype, like=A)
                if site < supp_A_min:
                    _support_A = [site, *_support_A]
                    _A = _kron(I, _A)
                else:  # site > actmax
                    _support_A = [*_support_A, site]
                    _A = _kron(_A, I)

        # reorder
        _A, _support_A = _reorder_kronecker_product(hilbert, _A, _support_A)
        _B, _support_B = _reorder_kronecker_product(hilbert, _B, _support_B)

        if len(_support_A) == len(_support_B) and np.array_equal(
            _support_A, _support_B
        ):
            # back to the case of non-intersecting with same support
            # TODO: Note: COO @ COO -> CSR
            return tuple(_support_A), _A @ _B
        else:
            raise ValueError("Something failed")



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
from collections.abc import Callable

import jax
from jax import numpy as jnp
from flax import struct

from netket.utils import timing
from netket.utils.types import PyTree


class _Uninitialized:
    pass


Uninitialized = _Uninitialized()

SolverT = Callable[["LinearOperator", PyTree], tuple[PyTree, Any]]
"""Signature for the solver used by a linear operator."""


@struct.dataclass
class LinearOperator:
    """
    S matrix base class.
    This can either be a jnp matrix, a lazy wrapper, or anything, as long as
    it satisfies this basic API.

    An AbstractSMatrix must support the following API:

    - :code:`__matmul__(y)`, meaning you must be able to do S@vec, where vec is
        either a PyTree of parameters or it's dense ravelling.
    - :code:`solve(y, **kwargs)`, which must solve the linear system Sx=y with
        any method available, usually but not necessarily defined inside the
        sr object stored inside the AbstractSMatrix.
        This function must accept arbitrary additional arguments.


    Additionally, you get for free:
    - :code:`__call__(y)`, meaning you must be able to do S(vec) = S@vec.
        This is defined by the base class so you don't need to define it.
        This guarantees that you can pass this matrix to sparse solvers.
    - :code:`to_dense(self)`, that will concretize the dense representation.
        You get a slow default fallback by default, you might specify a faster
        override.

    """

    diag_shift: float = 0.00
    """Diagonal shift added to the S matrix."""

    # PUBLIC API: METHOD TO EXTEND IF YOU WANT TO DEFINE A NEW S object
    def __matmul__(self, vec):
        raise NotImplementedError()

    def __add__(self, eps):
        return self.replace(diag_shift=self.diag_shift + eps)

    # PUBLIC API: METHOD TO EXTEND Optionally IF YOU WANT TO DEFINE A NEW S object with
    # custom logic
    def _solve(
        self, solve_fun: SolverT, y: PyTree, *, x0: PyTree | None = None, **kwargs
    ) -> PyTree:
        # dont pass x0 if it's unset.
        # some solvers might not need/require it.
        if x0 is not None:
            kwargs["x0"] = x0

        return solve_fun(self, y, **kwargs)

    # PUBLIC API: Extend _solve
    @timing.timed
    def solve(
        self, solve_fun: SolverT, y: PyTree, *, x0: PyTree | None = None, **kwargs
    ) -> PyTree:
        """
        Solve the linear system x=⟨S⟩⁻¹⟨y⟩ with the chosen iterative solver.

        Args:
            y: the vector y in the system above.
            kwargs: Any additional kwargs, which might or might not be used depending
                on the specific implementation.

        Returns:
            x: the PyTree solving the system.
            info: optional additional information provided by the solver. Might be
                None if there are no additional information provided.
        """
        return self._solve(jax.tree_util.Partial(solve_fun), y, x0=x0, **kwargs)

    # PUBLIC API: METHOD TO EXTEND IF YOU WANT TO DEFINE A NEW S object
    @timing.timed
    def to_dense(self) -> jnp.ndarray:
        """
        Convert the lazy matrix representation to a dense matrix representation.s

        Returns:
            A dense matrix representation of this S matrix.
        """
        raise NotImplementedError()

    # PUBLIC API: Only override if you want to, but there should be no need.
    def __call__(self, vec):
        return self @ vec

    # PUBLIC API: Only override if you want to, but there should be no need.
    def __array__(self) -> jnp.ndarray:
        return self.to_dense()

    def __post_init__(self):
        pass



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains logic to check version of NetKet's dependencies and hard-error in
case of outdated dependencies that we know might break NetKet's numerical results
silently or unexpectedly.
"""

from textwrap import dedent


from .version_check import module_version, version_string


def create_msg(pkg_name, cur_version, desired_version, extra_msg="", pip_pkg_name=None):
    if pip_pkg_name is None:  # pragma: no cover
        pip_pkg_name = pkg_name
    return dedent(
        f"""

        ##########################################################################################

        {pkg_name} version {cur_version} (< {desired_version}) is too old and incompatible with NetKet.
        Please update `{pkg_name}` by running the command:

            pip install --upgrade pip
            pip install --upgrade netket {pip_pkg_name}

        (assuming you are using pip. Similar commands can be used on conda).

        This error most likely happened because you either have an old version of `pip`
        or you are hard-coding the `{pkg_name}` version in a requirements file.

        {extra_msg}

        ##########################################################################################

        """
    )


if not module_version("jax") >= (0, 4, 16):  # pragma: no cover
    cur_version = version_string("jax")
    raise ImportError(create_msg("jax", cur_version, "0.4.16"))


if not module_version("optax") >= (0, 1, 3):  # pragma: no cover
    cur_version = version_string("optax")
    extra = """Reason: Optax is NetKet's provider of optimisers. Versions before 0.1.1 did not
               support complex numbers and silently returned wrong values, especially when
               using optimisers involving the norm of the gradient such as `Adam`.
               As recent versions of optax correctly work with complex numbers, please upgrade.
               """
    raise ImportError(create_msg("optax", cur_version, "0.1.3", extra))

if not module_version("flax") >= (0, 6, 5):  # pragma: no cover
    cur_version = version_string("flax")
    extra = """Reason: Flax is NetKet's default neural-network library. Versions before 0.5 had
               a bug and did not properly support complex numbers.
               """
    raise ImportError(create_msg("flax", cur_version, "0.6.5", extra))

if not module_version("plum") >= (2, 4, 0):  # pragma: no cover
    raise ImportError(
        create_msg(
            "plum", version_string("plum"), "2.4.0", pip_pkg_name="plum-dispatch"
        )
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from numba import jit

import numpy as np

from netket.operator import DiscreteOperator
from netket.utils import struct

from .base import MetropolisRule


@struct.dataclass
class HamiltonianRuleState:
    sections: np.ndarray
    """Preallocated array for sections"""


class HamiltonianRuleNumpy(MetropolisRule):
    r"""
    Rule for Numpy sampler backend proposing moves according to the terms in an operator.

    In this case, the transition matrix is taken to be:

    .. math::

       T( \mathbf{s} \rightarrow \mathbf{s}^\prime) =
        \frac{1}{\mathcal{N}(\mathbf{s})}\theta(|H_{\mathbf{s},\mathbf{s}^\prime}|),

    """

    operator: DiscreteOperator = struct.field(pytree_node=False)
    """The (hermitian) operator giving the transition amplitudes."""

    def __init__(self, operator: DiscreteOperator):
        """
        Constructs the Hamiltonian sampling rule for the
        :class:`netket.sampler.MetropolisNumpy` sampler.

        If you are using the standard jax sampler, look for
        :class:`netket.sampler.rules.HamiltonianRule` .

        Args:
            operator: The hermitian operator to be used to generate
                configurations.
        """
        if not isinstance(operator, DiscreteOperator):
            raise TypeError(
                "Argument to HamiltonianRule must be a valid operator, "
                f"but operator is a {type(operator)}."
            )
        self.operator = operator

    def init_state(rule, sampler, machine, params, key):
        if sampler.hilbert != rule.operator.hilbert:
            raise ValueError(
                f"""
            The hilbert space of the sampler ({sampler.hilbert}) and the hilbert space
            of the operator ({rule.operator.hilbert}) for HamiltonianRule must be the same.
            """
            )

        return HamiltonianRuleState(
            sections=np.empty(sampler.n_batches, dtype=np.int32)
        )

    def transition(rule, sampler, machine, parameters, state, rng, σ):
        σ = state.σ
        σ1 = state.σ1
        log_prob_corr = state.log_prob_corr

        sections = state.rule_state.sections
        σp = rule.operator.get_conn_flattened(σ, sections)[0]

        rand_vec = rng.uniform(0, 1, size=σ.shape[0])

        _choose(σp, sections, σ1, log_prob_corr, rand_vec)
        rule.operator.n_conn(σ1, sections)
        log_prob_corr -= np.log(sections)

    def __repr__(self):
        return f"HamiltonianRuleNumpy({self.operator})"


@jit(nopython=True)
def _choose(states, sections, out, w, rand_vec):
    low_range = 0
    for i, s in enumerate(sections):
        n_rand = low_range + int(np.floor(rand_vec[i] * (s - low_range)))
        out[i] = states[n_rand]
        w[i] = math.log(s - low_range)
        low_range = s



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .mpi import (
    mpi4jax_available as available,
    MPI,
    MPI_py_comm,
    MPI_jax_comm,
    n_nodes,
    node_number,
    rank,
)

from .primitives import (
    mpi_all,
    mpi_allgather,
    mpi_any,
    mpi_bcast,
    mpi_max,
    mpi_mean,
    mpi_sum,
)
from .primitives import (
    mpi_all_jax,
    mpi_allgather_jax,
    mpi_any_jax,
    mpi_bcast_jax,
    mpi_max_jax,
    mpi_mean_jax,
    mpi_sum_jax,
    mpi_gather,
    mpi_gather_jax,
    mpi_alltoall_jax,
    mpi_reduce_sum_jax,
    mpi_allreduce_sum_jax,
    mpi_scatter_jax,
)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import numpy as np

from netket.utils import StaticRange

from .homogeneous import HomogeneousHilbert
from .constraint import DiscreteHilbertConstraint, SumConstraint

FOCK_MAX = np.iinfo(np.intp).max - 1
"""
Maximum number of particles in the fock space.
It is `maxvalue(np.int64)-1` because we use N+1 in several formulas
and it would overflow.
"""


class Fock(HomogeneousHilbert):
    r"""Hilbert space obtained as tensor product of local fock basis."""

    def __init__(
        self,
        n_max: int | None = None,
        N: int = 1,
        n_particles: int | None = None,
        constraint: DiscreteHilbertConstraint | None = None,
    ):
        r"""
        Constructs a new ``Boson`` given a maximum occupation number, number of sites
        and total number of bosons.

        Args:
          n_max: Maximum occupation for a site (inclusive). If None, the local
            occupation number is unbounded.
          N: number of bosonic modes (default = 1)
          n_particles: Constraint for the number of particles. If None, no constraint
            is imposed.
          constraint: A custom constraint on the allowed configurations. This argument
            cannot be specified at the same time as :code:`n_particles`. The constraint
            must be a subclass of :class:`~netket.hilbert.constraint.DiscreteHilbertConstraint`

        Examples:
           Simple boson hilbert space.

           >>> from netket.hilbert import Fock
           >>> hi = Fock(n_max=5, n_particles=11, N=3)
           >>> print(hi.size)
           3
           >>> print(hi.n_states)
           15
        """
        self._n_max = n_max

        if n_particles is not None:
            if not isinstance(n_particles, int):
                raise TypeError(
                    f"n_particles must be an integer. Got {n_particles} ({type(n_particles)})"
                )
            if constraint is not None:
                raise ValueError(
                    "Cannot specify at the same time a total magnetization "
                    "constraint and a `custom_constraint."
                )

            n_particles = int(n_particles)
            if n_particles < 0:
                raise ValueError(
                    f"Number of particles must be >= 0, but received {n_particles}."
                )
            self._n_particles = n_particles

            if self._n_max is None:
                self._n_max = n_particles
            else:
                if self._n_max * N < n_particles:
                    raise Exception(
                        """The required total number of bosons is not compatible
                        with the given n_max."""
                    )
            constraint = SumConstraint(n_particles)
        else:
            self._n_particles = None

        if self._n_max is None:
            self._n_max = FOCK_MAX
        local_states = StaticRange(
            0,
            1,
            self._n_max + 1,
            dtype=np.int8 if self._n_max < 2**6 else int,  # type: ignore[arg-type]
        )

        super().__init__(local_states, N, constraint=constraint)

    @property
    def n_max(self) -> int | None:
        r"""The maximum number of bosons per site, or None
        if the number is unconstrained."""
        return self._n_max

    @property
    def n_particles(self) -> int | None:
        r"""The total number of particles, or None
        if the number is unconstrained."""
        return self._n_particles

    def __pow__(self, n) -> "Fock":
        if not self.constrained:
            return Fock(self.n_max, self.size * n)

        return NotImplemented

    def _mul_sametype_(self, other: "Fock") -> "Fock":
        assert type(self) == type(other)
        if self.n_max == other.n_max:
            if not self.constrained and not other.constrained:
                return Fock(self.n_max, N=self.size + other.size)

        return NotImplemented

    def ptrace(self, sites: int | list) -> Optional["Fock"]:
        if isinstance(sites, int):
            sites = [sites]

        for site in sites:
            if site < 0 or site >= self.size:
                raise ValueError(
                    f"Site {site} not in this hilbert space of site {self.size}"
                )

        if self.constrained:
            raise TypeError(
                "Cannot take the partial trace with a total particles constraint."
            )

        Nsites = len(sites)

        if self.size - Nsites == 0:
            return None
        else:
            return Fock(self.n_max, N=self.size - Nsites)

    def __repr__(self):
        if self.n_particles is not None:
            constraint = f", n_particles={self.n_particles}"
        elif self.constrained:
            constraint = f", {self._constraint}"
        else:
            constraint = ""

        nmax = self._n_max if self._n_max < FOCK_MAX else "FOCK_MAX"
        return f"Fock(n_max={nmax}{constraint}, N={self.size})"

    @property
    def _attrs(self):
        return (self.size, self._n_max, self.constraint)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from jax import numpy as jnp

import numpy as np

from netket import jax as nkjax
from netket.utils import config

from . import mean as _mean
from . import var as _var
from . import total_size as _total_size
from .mc_stats import Stats


def _get_blocks(data, block_size):
    chain_length = data.shape[1]

    n_blocks = int(np.floor(chain_length / float(block_size)))

    return data[:, 0 : n_blocks * block_size].reshape((-1, block_size)).mean(axis=1)


def _block_variance(data, l):
    blocks = _get_blocks(data, l)
    ts = _total_size(blocks)
    if ts > 0:
        return _var(blocks), ts
    else:
        return jnp.nan, 0


def _batch_variance(data):
    b_means = data.mean(axis=1)
    ts = _total_size(b_means)
    return _var(b_means), ts


# this is not batch_size maybe?
def statistics(data, batch_size=32):
    r"""
    Returns statistics of a given array (or matrix, see below) containing a stream of data.
    This is particularly useful to analyze Markov Chain data, but it can be used
    also for other type of time series.
    Assumes same shape on all MPI processes.

    Args:
        data (vector or matrix): The input data. It can be real or complex valued.
            * if a vector, it is assumed that this is a time series of data (not necessarily independent);
            * if a matrix, it is assumed that that rows :code:`data[i]` contain independent time series.

    Returns:
       Stats: A dictionary-compatible class containing the
             average (:code:`.mean`, :code:`["Mean"]`),
             variance (:code:`.variance`, :code:`["Variance"]`),
             the Monte Carlo standard error of the mean (:code:`error_of_mean`, :code:`["Sigma"]`),
             an estimate of the autocorrelation time (:code:`tau_corr`, :code:`["TauCorr"]`), and the
             Gelman-Rubin split-Rhat diagnostic (:code:`.R_hat`, :code:`["R_hat"]`).

             These properties can be accessed both the attribute and the dictionary-style syntax
             (both indicated above).

             The split-Rhat diagnostic is based on comparing intra-chain and inter-chain
             statistics of the sample and is thus only available for 2d-array inputs where
             the rows are independently sampled MCMC chains. In an ideal MCMC samples,
             R_hat should be 1.0. If it deviates from this value too much, this indicates
             MCMC convergence issues. Thresholds such as R_hat > 1.1 or even R_hat > 1.01 have
             been suggested in the literature for when to discard a sample. (See, e.g.,
             Gelman et al., `Bayesian Data Analysis <http://www.stat.columbia.edu/~gelman/book/>`_,
             or Vehtari et al., `arXiv:1903.08008 <https://arxiv.org/abs/1903.08008>`_.)
    """
    return _statistics(data, batch_size)


@partial(jax.jit, static_argnums=1)
def _statistics(data, batch_size):
    data = jnp.atleast_1d(data)
    if data.ndim == 1:
        data = data.reshape((1, -1))

    if data.ndim > 2:
        raise NotImplementedError("Statistics are implemented only for ndim<=2")

    mean = _mean(data)
    variance = _var(data)

    ts = _total_size(data)

    bare_var = variance

    batch_var, n_batches = _batch_variance(data)

    l_block = max(1, data.shape[1] // batch_size)

    block_var, n_blocks = _block_variance(data, l_block)

    tau_batch = ((ts / n_batches) * batch_var / bare_var - 1) * 0.5
    tau_block = ((ts / n_blocks) * block_var / bare_var - 1) * 0.5

    batch_good = (tau_batch < 6 * data.shape[1]) * (n_batches >= batch_size)
    block_good = (tau_block < 6 * l_block) * (n_blocks >= batch_size)

    stat_dtype = nkjax.dtype_real(data.dtype)

    # if batch_good:
    #    error_of_mean = jnp.sqrt(batch_var / n_batches)
    #    tau_corr = jnp.max(0, tau_batch)
    # elif block_good:
    #    error_of_mean = jnp.sqrt(block_var / n_blocks)
    #    tau_corr = jnp.max(0, tau_block)
    # else:
    #    error_of_mean = jnp.nan
    #    tau_corr = jnp.nan
    # jax style

    def batch_good_err(args):
        batch_var, tau_batch, *_ = args
        error_of_mean = jnp.sqrt(batch_var / n_batches)
        tau_corr = jnp.clip(tau_batch, 0)
        return jnp.asarray(error_of_mean, dtype=stat_dtype), jnp.asarray(
            tau_corr, dtype=stat_dtype
        )

    def block_good_err(args):
        _, _, block_var, tau_block = args
        error_of_mean = jnp.sqrt(block_var / n_blocks)
        tau_corr = jnp.clip(tau_block, 0)
        return jnp.asarray(error_of_mean, dtype=stat_dtype), jnp.asarray(
            tau_corr, dtype=stat_dtype
        )

    def nan_err(args):
        return jnp.asarray(jnp.nan, dtype=stat_dtype), jnp.asarray(
            jnp.nan, dtype=stat_dtype
        )

    def batch_not_good(args):
        batch_var, tau_batch, block_var, tau_block, block_good = args
        return jax.lax.cond(
            block_good,
            block_good_err,
            nan_err,
            (batch_var, tau_batch, block_var, tau_block),
        )

    error_of_mean, tau_corr = jax.lax.cond(
        batch_good,
        batch_good_err,
        batch_not_good,
        (batch_var, tau_batch, block_var, tau_block, block_good),
    )

    if n_batches > 1:
        N = data.shape[-1]

        if not config.netket_use_plain_rhat:
            # compute split-chain batch variance
            local_batch_size = data.shape[0]
            if N % 2 == 0:
                # split each chain in the middle,
                # like [[1 2 3 4]] -> [[1 2][3 4]]
                batch_var, _ = _batch_variance(
                    data.reshape(2 * local_batch_size, N // 2)
                )
            else:
                # drop the last sample of each chain for an even split,
                # like [[1 2 3 4 5]] -> [[1 2][3 4]]
                batch_var, _ = _batch_variance(
                    data[:, :-1].reshape(2 * local_batch_size, N // 2)
                )

        # V_loc = _np.var(data, axis=-1, ddof=0)
        # W_loc = _np.mean(V_loc)
        # W = _mean(W_loc)
        # # This approximation seems to hold well enough for larger n_samples
        W = variance

        R_hat = jnp.sqrt((N - 1) / N + batch_var / W)
    else:
        R_hat = jnp.nan

    res = Stats(mean, error_of_mean, variance, tau_corr, R_hat)

    return res



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from collections.abc import Callable
import warnings

import jax
from jax import numpy as jnp

from netket import jax as nkjax
from netket.stats import Stats
from netket.utils.types import PyTree
from netket.utils.dispatch import dispatch

from netket.operator import (
    AbstractOperator,
    DiscreteOperator,
    DiscreteJaxOperator,
    ContinuousOperator,
    Squared,
)

# to move up once stabilized
from netket.operator._abstract_observable import AbstractObservable

from netket.vqs import expect

from netket.vqs.mc import (
    kernels,
    get_local_kernel,
    get_local_kernel_arguments,
)

from .state import MCState


# Dispatches to select what expect-kernel to use
@dispatch
def get_local_kernel(vstate: MCState, Ô: Squared, chunk_size: int):  # noqa: F811
    return kernels.local_value_squared_kernel_chunked


@dispatch
def get_local_kernel(  # noqa: F811
    vstate: MCState, Ô: DiscreteJaxOperator, chunk_size: int
):  # noqa: F811
    return kernels.local_value_kernel_jax_chunked


@dispatch
def get_local_kernel(  # noqa: F811
    vstate: MCState, Ô: DiscreteOperator, chunk_size: int
):
    return kernels.local_value_kernel_chunked


def _local_continuous_kernel(kernel, logpsi, pars, σ, args, *, chunk_size=None):
    def _kernel(σ):
        return kernel(logpsi, pars, σ, args)

    return nkjax.apply_chunked(_kernel, in_axes=0, chunk_size=chunk_size)(σ)


@dispatch
def get_local_kernel(  # noqa: F811
    vstate: MCState, Ô: ContinuousOperator, chunk_size: int
):
    return nkjax.HashablePartial(_local_continuous_kernel, Ô._expect_kernel)


# If batch_size is unspecified, set it to None
@expect.dispatch
def expect_chunking_unspecified(vstate: MCState, operator: AbstractObservable):
    return expect(vstate, operator, None)


# if no implementation exists for batched, fall back to unbatched methods.
@expect.dispatch(precedence=-10)
def expect_fallback(
    vstate: MCState, operator: AbstractObservable, chunk_size: int | tuple
):  # noqa: F811
    warnings.warn(
        f"Ignoring chunk_size={chunk_size} for expect_and_grad method with signature "
        f"({type(vstate)}, {type(operator)}) because no implementation supporting "
        f"chunking for this signature exists."
    )

    return expect(vstate, operator, None)


@expect.dispatch
def expect_mcstate_operator_chunked(
    vstate: MCState, Ô: AbstractOperator, chunk_size: int
) -> Stats:  # noqa: F811
    σ, args = get_local_kernel_arguments(vstate, Ô)

    local_estimator_fun = get_local_kernel(vstate, Ô, chunk_size)

    return _expect_chunking(
        chunk_size,
        local_estimator_fun,
        vstate._apply_fun,
        vstate.sampler.machine_pow,
        vstate.parameters,
        vstate.model_state,
        σ,
        args,
    )


@partial(jax.jit, static_argnums=(0, 1, 2))
def _expect_chunking(
    chunk_size: int,
    local_value_kernel: Callable,
    model_apply_fun: Callable,
    machine_pow: int,
    parameters: PyTree,
    model_state: PyTree,
    σ: jnp.ndarray,
    args: PyTree,
) -> Stats:
    σ_shape = σ.shape

    if jnp.ndim(σ) != 2:
        σ = σ.reshape((-1, σ_shape[-1]))

    def logpsi(w, σ):
        return model_apply_fun({"params": w, **model_state}, σ)

    def log_pdf(w, σ):
        return machine_pow * model_apply_fun({"params": w, **model_state}, σ).real

    _, Ō_stats = nkjax.expect(
        log_pdf,
        partial(local_value_kernel, logpsi, chunk_size=chunk_size),
        parameters,
        σ,
        args,
        n_chains=σ_shape[0],
    )

    return Ō_stats



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from collections.abc import Callable

import jax
from jax import numpy as jnp
from flax.core.scope import CollectionFilter, DenyList  # noqa: F401

from netket import jax as nkjax
from netket.stats import Stats, statistics
from netket.utils import mpi
from netket.utils.types import PyTree
from netket.utils.dispatch import dispatch

from netket.operator import (
    AbstractOperator,
)

from netket.vqs.mc import (
    get_local_kernel_arguments,
    get_local_kernel,
)

from .state import MCState


@dispatch
def expect_and_forces(  # noqa: F811
    vstate: MCState,
    Ô: AbstractOperator,
    chunk_size: None,
    *,
    mutable: CollectionFilter = False,
) -> tuple[Stats, PyTree]:
    σ, args = get_local_kernel_arguments(vstate, Ô)

    local_estimator_fun = get_local_kernel(vstate, Ô)

    Ō, Ō_grad, new_model_state = forces_expect_hermitian(
        local_estimator_fun,
        vstate._apply_fun,
        mutable,
        vstate.parameters,
        vstate.model_state,
        σ,
        args,
    )

    if mutable is not False:
        vstate.model_state = new_model_state

    return Ō, Ō_grad


@partial(jax.jit, static_argnums=(0, 1, 2))
def forces_expect_hermitian(
    local_value_kernel: Callable,
    model_apply_fun: Callable,
    mutable: CollectionFilter,
    parameters: PyTree,
    model_state: PyTree,
    σ: jnp.ndarray,
    local_value_args: PyTree,
) -> tuple[PyTree, PyTree]:
    n_chains = σ.shape[0]
    if σ.ndim >= 3:
        σ = jax.lax.collapse(σ, 0, 2)

    n_samples = σ.shape[0] * mpi.n_nodes

    O_loc = local_value_kernel(
        model_apply_fun,
        {"params": parameters, **model_state},
        σ,
        local_value_args,
    )

    Ō = statistics(O_loc.reshape((n_chains, -1)))

    O_loc -= Ō.mean

    # Then compute the vjp.
    # Code is a bit more complex than a standard one because we support
    # mutable state (if it's there)
    is_mutable = mutable is not False
    _, vjp_fun, *new_model_state = nkjax.vjp(
        lambda w: model_apply_fun({"params": w, **model_state}, σ, mutable=mutable),
        parameters,
        conjugate=True,
        has_aux=is_mutable,
    )
    Ō_grad = vjp_fun(jnp.conjugate(O_loc) / n_samples)[0]

    Ō_grad, _ = mpi.mpi_sum_jax(Ō_grad)

    new_model_state = new_model_state[0] if is_mutable else None

    return (
        Ō,
        Ō_grad,
        new_model_state,
    )



# Copyright 2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp

from netket.utils import struct
from netket.utils.types import Array

from .base import HilbertIndex, is_indexable


@struct.dataclass
class UniformTensorProductHilbertIndex(HilbertIndex):
    """
    Indexes a tensor product of multiple HilbertIndexes, ordered in LSB order
    (so when increasing by one the index, the right-most degree of freedom is increased).
    This index assumes a tensor product of the same Hilbert index with itself _size times,
    and is used to represent all UniformHilbert spaces.
    """

    # tensor product with uniform local space
    local_index: HilbertIndex
    size: int = struct.field(pytree_node=False)

    @property
    def local_size(self) -> int:
        return self.local_index.n_states

    @property
    def n_states(self) -> int:
        return self.local_size**self.size

    @property
    def local_states(self) -> Array:
        return self.local_index.all_states()

    @property
    def _basis(self) -> Array:
        return self.local_size ** jax.lax.iota(jnp.int32, self.size)[::-1]

    @jax.jit
    def states_to_numbers(self, states: Array) -> Array:
        local_numbers = self.local_index.states_to_numbers(states, dtype=jnp.int32)
        return local_numbers @ self._basis

    @jax.jit
    def numbers_to_states(self, numbers: Array) -> Array:
        local_numbers = (numbers[..., None] // self._basis) % self.local_size
        return self.local_index.numbers_to_states(local_numbers)

    @jax.jit
    def all_states(self) -> Array:
        return self.numbers_to_states(jnp.arange(self.n_states, dtype=jnp.int32))

    @property
    def is_indexable(self) -> bool:
        return is_indexable(self.n_states)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
import jax.flatten_util
import jax.numpy as jnp

import numpy as np

from netket import jax as nkjax
from netket.utils import wrap_to_support_scalar
from netket.utils.types import Array, Callable, PyTree


def jacobian_real_holo(forward_fn: Callable, params: PyTree, samples: Array) -> PyTree:
    """Calculates Jacobian entries by vmapping grad.
    Assumes the function is R→R or holomorphic C→C, so single grad is enough

    Args:
        forward_fn: the log wavefunction ln Ψ
        params : a pytree of parameters p
        samples : an array of n samples σ

    Returns:
        The Jacobian matrix ∂/∂pₖ ln Ψ(σⱼ) as a PyTree
    """

    y, vjp_fun = jax.vjp(
        lambda pars: wrap_to_support_scalar(forward_fn)(pars, samples), params
    )
    (res,) = vjp_fun(np.array(1.0, dtype=jnp.result_type(y)))
    return res


def _jacobian_cplx(
    forward_fn: Callable,
    params: PyTree,
    samples: Array,
    *,
    _build_fn: Callable = partial(jax.tree_util.tree_map, jax.lax.complex),
) -> PyTree:
    """Calculates Jacobian entries by vmapping grad.
    Assumes the function is R→C, backpropagates 1 and -1j

    Args:
        forward_fn: the log wavefunction ln Ψ
        params : a pytree of parameters p
        samples : an array of n samples σ

    Returns:
        The Jacobian matrix ∂/∂pₖ ln Ψ(σⱼ) as a PyTree
    """

    y, vjp_fun = jax.vjp(
        lambda pars: wrap_to_support_scalar(forward_fn)(pars, samples), params
    )
    if not jnp.issubdtype(jnp.result_type(y), jnp.complexfloating):
        raise TypeError("Cannot build the complex jacobian for a real-valued function.")

    (gr,) = vjp_fun(np.array(1.0, dtype=jnp.result_type(y)))
    (gi,) = vjp_fun(np.array(-1.0j, dtype=jnp.result_type(y)))
    return _build_fn(gr, gi)


jacobian_cplx = partial(_jacobian_cplx, _build_fn=lambda *x: x)


def stack_jacobian_tuple(ok_re_im):
    """
    stack the real and imaginary parts of ΔOⱼₖ along the sample axis

    Re[S] = Re[(ΔOᵣ + i ΔOᵢ)ᴴ(ΔOᵣ + i ΔOᵢ)] = ΔOᵣᵀ ΔOᵣ + ΔOᵢᵀ ΔOᵢ = [ΔOᵣ ΔOᵢ]ᵀ [ΔOᵣ ΔOᵢ]

    Args:
        centered_oks_re_im : a tuple (ΔOᵣ, ΔOᵢ) of two PyTrees representing the real and imag part of ΔOⱼₖ
    """
    re, im = ok_re_im
    return jax.tree_util.tree_map(lambda re, im: jnp.stack([re, im], axis=0), re, im)


jacobian_real_holo_fun = jacobian_real_holo
jacobian_cplx_fun = nkjax.compose(stack_jacobian_tuple, jacobian_cplx)
